"""Module contains the class to create a fuzzy prompt."""
import asyncio
import math
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    List,
    Optional,
    Tuple,
    Union,
    cast,
)

from pfzy import fuzzy_match
from pfzy.score import fzy_scorer, substr_scorer
from pfzy.types import HAYSTACKS
from prompt_toolkit.application.application import Application
from prompt_toolkit.buffer import Buffer
from prompt_toolkit.filters.cli import IsDone
from prompt_toolkit.layout.containers import (
    ConditionalContainer,
    FloatContainer,
    HSplit,
    Window,
)
from prompt_toolkit.layout.controls import BufferControl, DummyControl
from prompt_toolkit.layout.dimension import Dimension, LayoutDimension
from prompt_toolkit.layout.layout import Layout
from prompt_toolkit.layout.processors import AfterInput, BeforeInput
from prompt_toolkit.lexers.base import SimpleLexer
from prompt_toolkit.validation import ValidationError
from prompt_toolkit.widgets.base import Frame

from InquirerPy.base import FakeDocument, InquirerPyUIListControl
from InquirerPy.base.list import BaseListPrompt
from InquirerPy.containers.instruction import InstructionWindow
from InquirerPy.containers.message import MessageWindow
from InquirerPy.containers.validation import ValidationFloat
from InquirerPy.enum import INQUIRERPY_POINTER_SEQUENCE
from InquirerPy.exceptions import InvalidArgument
from InquirerPy.separator import Separator
from InquirerPy.utils import (
    InquirerPyDefault,
    InquirerPyKeybindings,
    InquirerPyListChoices,
    InquirerPyMessage,
    InquirerPySessionResult,
    InquirerPyStyle,
    InquirerPyValidate,
    calculate_height,
)

if TYPE_CHECKING:
    from prompt_toolkit.key_binding.key_processor import KeyPressEvent

__all__ = ["FuzzyPrompt"]


class InquirerPyFuzzyControl(InquirerPyUIListControl):
    """An :class:`~prompt_toolkit.layout.UIControl` class that displays a list of choices.

    This only displays the chocies. The actual input buffer will be handled by a separate
    :class:`~prompt_toolkit.layout.BufferControl`.

    Reference the parameter definition in :class:`.FuzzyPrompt`.
    """

    def __init__(
        self,
        choices: InquirerPyListChoices,
        pointer: str,
        marker: str,
        current_text: Callable[[], str],
        max_lines: int,
        session_result: Optional[InquirerPySessionResult],
        multiselect: bool,
        marker_pl: str,
        match_exact: bool,
    ) -> None:
        self._pointer = pointer
        self._marker = marker
        self._marker_pl = marker_pl
        self._current_text = current_text
        self._max_lines = max_lines if max_lines > 0 else 1
        self._scorer = fzy_scorer if not match_exact else substr_scorer
        super().__init__(
            choices=choices,
            default=None,
            session_result=session_result,
            multiselect=multiselect,
        )

    def _format_choices(self) -> None:
        for index, choice in enumerate(self.choices):
            if isinstance(choice["value"], Separator):
                raise InvalidArgument(
                    "fuzzy prompt argument choices should not contain Separator"
                )
            choice["index"] = index
            choice["indices"] = []
        self._filtered_choices = self.choices
        self._first_line = 0
        self._last_line = min(self._max_lines, self.choice_count)
        self._height = self._last_line - self._first_line

    def _get_hover_text(self, choice) -> List[Tuple[str, str]]:
        """Get the current highlighted line of text.

        If in the middle of filtering, loop through the char and color
        indices matched char into style class `class:fuzzy_match`.

        Returns:
            FormattedText in list of tuple format.
        """
        display_choices = []
        display_choices.append(("class:pointer", self._pointer))
        display_choices.append(
            (
                "class:marker",
                self._marker
                if self.choices[choice["index"]]["enabled"]
                else self._marker_pl,
            )
        )
        display_choices.append(("[SetCursorPosition]", ""))
        if not choice["indices"]:
            display_choices.append(("class:pointer", choice["name"]))
        else:
            indices = set(choice["indices"])
            for index, char in enumerate(choice["name"]):
                if index in indices:
                    display_choices.append(("class:fuzzy_match", char))
                else:
                    display_choices.append(("class:pointer", char))
        return display_choices

    def _get_normal_text(self, choice) -> List[Tuple[str, str]]:
        """Get the line of text in `FormattedText`.

        If in the middle of filtering, loop through the char and color
        indices matched char into `class:fuzzy_match`.

        Calculate spaces of pointer to make the choice equally align.

        Returns:
            FormattedText in list of tuple format.
        """
        display_choices = []
        display_choices.append(("class:pointer", len(self._pointer) * " "))
        display_choices.append(
            (
                "class:marker",
                self._marker
                if self.choices[choice["index"]]["enabled"]
                else self._marker_pl,
            )
        )
        if not choice["indices"]:
            display_choices.append(("", choice["name"]))
        else:
            indices = set(choice["indices"])
            for index, char in enumerate(choice["name"]):
                if index in indices:
                    display_choices.append(("class:fuzzy_match", char))
                else:
                    display_choices.append(("", char))
        return display_choices

    def _get_formatted_choices(self) -> List[Tuple[str, str]]:
        """Get all available choices in formatted text format.

        Overriding this method because `self.choice` will be the
        full choice list. Using `self.filtered_choice` to get
        a list of choice based on current_text.

        Returns:
            FormattedText in list of tuple format.
        """
        display_choices = []
        if self.choice_count == 0:
            self._selected_choice_index = 0
            return display_choices

        if self._selected_choice_index < 0:
            self._selected_choice_index = 0
        elif self._selected_choice_index >= self.choice_count:
            self._selected_choice_index = self.choice_count - 1

        if (self._last_line - self._first_line) < min(self.choice_count, self._height):
            self._last_line = min(self.choice_count, self._height)
            self._first_line = self._last_line - min(self.choice_count, self._height)

        if self._selected_choice_index <= self._first_line:
            self._first_line = self._selected_choice_index
            self._last_line = self._first_line + min(self._height, self.choice_count)
        elif self._selected_choice_index >= self._last_line:
            self._last_line = self._selected_choice_index + 1
            self._first_line = self._last_line - min(self._height, self.choice_count)

        if self._last_line > self.choice_count:
            self._last_line = self.choice_count
            self._first_line = self._last_line - min(self._height, self.choice_count)
        if self._first_line < 0:
            self._first_line = 0
            self._last_line = self._first_line + min(self._height, self.choice_count)

        for index in range(self._first_line, self._last_line):
            if index == self.selected_choice_index:
                display_choices += self._get_hover_text(self._filtered_choices[index])
            else:
                display_choices += self._get_normal_text(self._filtered_choices[index])
            display_choices.append(("", "\n"))
        if display_choices:
            display_choices.pop()
        return display_choices

    async def _filter_choices(self, wait_time: float) -> List[Dict[str, Any]]:
        """Call to filter choices using fzy fuzzy match.

        Args:
            wait_time: Additional time to wait before filtering the choice.

        Returns:
            Filtered choices.
        """
        if not self._current_text():
            for choice in self.choices:
                choice["indices"] = []
            choices = self.choices
        else:
            await asyncio.sleep(wait_time)
            choices = await fuzzy_match(
                self._current_text(),
                cast(HAYSTACKS, self.choices),
                key="name",
                scorer=self._scorer,
            )
        return choices

    @property
    def selection(self) -> Dict[str, Any]:
        """Override this value since `self.choice` does not indicate the choice displayed.

        `self.filtered_choice` is the up to date choice displayed.

        Returns:
            A dictionary of name and value for the current pointed choice.
        """
        return self._filtered_choices[self.selected_choice_index]

    @property
    def choice_count(self) -> int:
        """int: Filtered choice count."""
        return len(self._filtered_choices)


class FuzzyPrompt(BaseListPrompt):
    """Create a prompt that lists choices while also allowing fuzzy search like fzf.

    A wrapper class around :class:`~prompt_toolkit.application.Application`.

    Fuzzy search using :func:`pfzy.match.fuzzy_match` function.

    Override the default keybindings for up/down as j/k cannot be bind even if `editing_mode` is vim
    due to the input buffer.

    Args:
        message: The question to ask the user.
            Refer to :ref:`pages/dynamic:message` documentation for more details.
        choices: List of choices to display and select.
            Refer to :ref:`pages/dynamic:choices` documentation for more details.
        style: An :class:`InquirerPyStyle` instance.
            Refer to :ref:`Style <pages/style:Alternate Syntax>` documentation for more details.
        vi_mode: Use vim keybinding for the prompt.
            Refer to :ref:`pages/kb:Keybindings` documentation for more details.
        default: Set the default value in the search buffer.
            Different than other list type prompts, the `default` parameter tries to replicate what fzf does and
            add the value in `default` to search buffer so it starts searching immediatelly.
            Refer to :ref:`pages/dynamic:default` documentation for more details.
        qmark: Question mark symbol. Custom symbol that will be displayed infront of the question before its answered.
        amark: Answer mark symbol. Custom symbol that will be displayed infront of the question after its answered.
        pointer: Pointer symbol. Customer symbol that will be used to indicate the current choice selection.
        instruction: Short instruction to display next to the question.
        long_instruction: Long instructions to display at the bottom of the prompt.
        validate: Add validation to user input.
            The main use case for this prompt would be when `multiselect` is True, you can enforce a min/max selection.
            Refer to :ref:`pages/validator:Validator` documentation for more details.
        invalid_message: Error message to display when user input is invalid.
            Refer to :ref:`pages/validator:Validator` documentation for more details.
        transformer: A function which performs additional transformation on the value that gets printed to the terminal.
            Different than `filter` parameter, this is only visual effect and won’t affect the actual value returned by :meth:`~InquirerPy.base.simple.BaseSimplePrompt.execute`.
            Refer to :ref:`pages/dynamic:transformer` documentation for more details.
        filter: A function which performs additional transformation on the result.
            This affects the actual value returned by :meth:`~InquirerPy.base.simple.BaseSimplePrompt.execute`.
            Refer to :ref:`pages/dynamic:filter` documentation for more details.
        height: Preferred height of the prompt.
            Refer to :ref:`pages/height:Height` documentation for more details.
        max_height: Max height of the prompt.
            Refer to :ref:`pages/height:Height` documentation for more details.
        multiselect: Enable multi-selection on choices.
            You can use `validate` parameter to control min/max selections.
            Setting to True will also change the result from a single value to a list of values.
        prompt: Input prompt symbol. Custom symbol to display infront of the input buffer to indicate for input.
        border: Create border around the choice window.
        info: Display choice information similar to fzf --info=inline next to the prompt.
        match_exact: Use exact sub-string match instead of using fzy fuzzy match algorithm.
        exact_symbol: Custom symbol to display in the info section when `info=True`.
        marker: Marker Symbol. Custom symbol to indicate if a choice is selected.
            This will take effects when `multiselect` is True.
        marker_pl: Marker place holder when the choice is not selected.
            This is empty space by default.
        keybindings: Customise the builtin keybindings.
            Refer to :ref:`pages/kb:Keybindings` for more details.
        cycle: Return to top item if hit bottom during navigation or vice versa.
        wrap_lines: Soft wrap question lines when question exceeds the terminal width.
        raise_keyboard_interrupt: Raise the :class:`KeyboardInterrupt` exception when `ctrl-c` is pressed. If false, the result
            will be `None` and the question is skiped.
        mandatory: Indicate if the prompt is mandatory. If True, then the question cannot be skipped.
        mandatory_message: Error message to show when user attempts to skip mandatory prompt.
        session_result: Used internally for :ref:`index:Classic Syntax (PyInquirer)`.

    Examples:
        >>> from InquirerPy import inquirer
        >>> result = inquirer.fuzzy(message="Select one:", choices=[1, 2, 3]).execute()
        >>> print(result)
        1
    """

    def __init__(
        self,
        message: InquirerPyMessage,
        choices: InquirerPyListChoices,
        default: InquirerPyDefault = "",
        pointer: str = INQUIRERPY_POINTER_SEQUENCE,
        style: Optional[InquirerPyStyle] = None,
        vi_mode: bool = False,
        qmark: str = "?",
        amark: str = "?",
        transformer: Optional[Callable[[Any], Any]] = None,
        filter: Optional[Callable[[Any], Any]] = None,
        instruction: str = "",
        long_instruction: str = "",
        multiselect: bool = False,
        prompt: str = INQUIRERPY_POINTER_SEQUENCE,
        marker: str = INQUIRERPY_POINTER_SEQUENCE,
        marker_pl: str = " ",
        border: bool = False,
        info: bool = True,
        match_exact: bool = False,
        exact_symbol: str = " E",
        height: Optional[Union[str, int]] = None,
        max_height: Optional[Union[str, int]] = None,
        validate: Optional[InquirerPyValidate] = None,
        invalid_message: str = "Invalid input",
        keybindings: Optional[InquirerPyKeybindings] = None,
        cycle: bool = True,
        wrap_lines: bool = True,
        raise_keyboard_interrupt: bool = True,
        mandatory: bool = True,
        mandatory_message: str = "Mandatory prompt",
        session_result: Optional[InquirerPySessionResult] = None,
    ) -> None:
        if not keybindings:
            keybindings = {}
        self._prompt = prompt
        self._info = info
        self._task = None
        self._rendered = False
        self._exact_symbol = exact_symbol

        keybindings = {
            "up": [{"key": "up"}, {"key": "c-p"}],
            "down": [{"key": "down"}, {"key": "c-n"}],
            "toggle": [],
            "toggle-exact": [],
            **keybindings,
        }
        super().__init__(
            message=message,
            style=style,
            border=border,
            vi_mode=vi_mode,
            qmark=qmark,
            amark=amark,
            transformer=transformer,
            filter=filter,
            validate=validate,
            invalid_message=invalid_message,
            multiselect=multiselect,
            instruction=instruction,
            long_instruction=long_instruction,
            keybindings=keybindings,
            cycle=cycle,
            wrap_lines=wrap_lines,
            raise_keyboard_interrupt=raise_keyboard_interrupt,
            mandatory=mandatory,
            mandatory_message=mandatory_message,
            session_result=session_result,
        )
        self.kb_func_lookup = {"toggle-exact": [{"func": self._toggle_exact}]}
        self._default = (
            default
            if not isinstance(default, Callable)
            else cast(Callable, default)(self._result)
        )
        self._height_offset += 1  # search input
        self._dimmension_height, self._dimmension_max_height = calculate_height(
            height, max_height, height_offset=self.height_offset
        )

        self._content_control: InquirerPyFuzzyControl = InquirerPyFuzzyControl(
            choices=choices,
            pointer=pointer,
            marker=marker,
            current_text=self._get_current_text,
            max_lines=self._dimmension_max_height,
            session_result=session_result,
            multiselect=multiselect,
            marker_pl=marker_pl,
            match_exact=match_exact,
        )

        self._buffer = Buffer(on_text_changed=self._on_text_changed)
        input_window = Window(
            height=LayoutDimension.exact(1),
            content=BufferControl(
                self._buffer,
                [
                    AfterInput(self._generate_after_input),
                    BeforeInput(self._generate_before_input),
                ],
                lexer=SimpleLexer("class:input"),
            ),
        )

        choice_height_dimmension = lambda: Dimension(
            max=self._dimmension_max_height,
            preferred=self._dimmension_height,
            min=self.content_control._height if self.content_control._height > 0 else 1,
        )
        self.choice_window = Window(
            content=self.content_control,
            height=choice_height_dimmension,
            dont_extend_height=True,
        )

        main_content_window = HSplit([input_window, self.choice_window])
        if self._border:
            main_content_window = Frame(main_content_window)
        self._layout = Layout(
            FloatContainer(
                content=HSplit(
                    [
                        MessageWindow(
                            message=self._get_prompt_message,
                            filter=True,
                            wrap_lines=self._wrap_lines,
                            show_cursor=True,
                        ),
                        ConditionalContainer(
                            main_content_window,
                            filter=~IsDone(),
                        ),
                        ConditionalContainer(
                            Window(content=DummyControl()),
                            filter=~IsDone() & self._is_displaying_long_instruction,
                        ),
                        InstructionWindow(
                            message=self._long_instruction,
                            filter=~IsDone() & self._is_displaying_long_instruction,
                            wrap_lines=self._wrap_lines,
                        ),
                    ],
                ),
                floats=[
                    ValidationFloat(
                        invalid_message=self._get_error_message,
                        filter=self._is_invalid & ~IsDone(),
                        wrap_lines=self._wrap_lines,
                        left=0,
                        bottom=self._validation_window_bottom_offset,
                    ),
                ],
            )
        )
        self._layout.focus(input_window)

        self._application = Application(
            layout=self._layout,
            style=self._style,
            key_bindings=self._kb,
            editing_mode=self._editing_mode,
            after_render=self._after_render,
        )

    def _toggle_exact(self, _, value: Optional[bool] = None) -> None:
        """Toggle matching algorithm.

        Switch between fzy fuzzy match or sub-string exact match.

        Args:
            value: Specify the value to toggle.
        """
        if value is not None:
            self.content_control._scorer = fzy_scorer if not value else substr_scorer
        else:
            self.content_control._scorer = (
                fzy_scorer
                if self.content_control._scorer == substr_scorer
                else substr_scorer
            )

    def _on_rendered(self, _) -> None:
        """Render callable choices and set the buffer default text.

        Setting buffer default text has to be after application is rendered and choice are loaded,
        because `self._filter_choices` will use the event loop from `Application`.
        """
        if self._default:
            default_text = str(self._default)
            self._buffer.text = default_text
            self._buffer.cursor_position = len(default_text)

    def _handle_toggle_all(self, _, value: Optional[bool] = None) -> None:
        """Toggle all choice `enabled` status.

        Args:
            value: Specify the value to toggle.
        """
        if not self._multiselect:
            return
        for choice in self.content_control._filtered_choices:
            raw_choice = self.content_control.choices[choice["index"]]
            if isinstance(raw_choice["value"], Separator):
                continue
            raw_choice["enabled"] = value if value else not raw_choice["enabled"]

    def _generate_after_input(self) -> List[Tuple[str, str]]:
        """Virtual text displayed after the user input."""
        display_message = []
        if self._info:
            display_message.append(("", "  "))
            display_message.append(
                (
                    "class:fuzzy_info",
                    f"{self.content_control.choice_count}/{len(self.content_control.choices)}",
                )
            )
            if self._multiselect:
                display_message.append(
                    ("class:fuzzy_info", f" ({len(self.selected_choices)})")
                )
            if self.content_control._scorer == substr_scorer:
                display_message.append(("class:fuzzy_info", self._exact_symbol))
        return display_message

    def _generate_before_input(self) -> List[Tuple[str, str]]:
        """Display prompt symbol as virtual text before user input."""
        display_message = []
        display_message.append(("class:fuzzy_prompt", "%s " % self._prompt))
        return display_message

    def _filter_callback(self, task):
        """Redraw `self._application` when the filter task is finished."""
        if task.cancelled():
            return
        self.content_control._filtered_choices = task.result()
        self._application.invalidate()

    def _calculate_wait_time(self) -> float:
        """Calculate wait time to smoother the application on big data set.

        Using digit of the choices lengeth to get wait time.
        For digit greater than 6, using formula 2^(digit - 5) * 0.3 to increase the wait_time.

        Returns:
            Desired wait time before running the filter.
        """
        wait_table = {
            2: 0.05,
            3: 0.1,
            4: 0.2,
            5: 0.3,
        }
        digit = 1
        if len(self.content_control.choices) > 0:
            digit = int(math.log10(len(self.content_control.choices))) + 1

        if digit < 2:
            return 0.0
        if digit in wait_table:
            return wait_table[digit]
        return wait_table[5] * (2 ** (digit - 5))

    def _on_text_changed(self, _) -> None:
        """Handle buffer text change event.

        1. Check if there is current task running.
        2. Cancel if already has task, increase wait_time
        3. Create a filtered_choice task in asyncio event loop
        4. Add callback

        1. Run a new filter on all choices.
        2. Re-calculate current selected_choice_index
            if it exceeds the total filtered_choice.
        3. Avoid selected_choice_index less than zero,
            this fix the issue of cursor lose when:
            choice -> empty choice -> choice

        Don't need to create or check asyncio event loop, `prompt_toolkit`
        application already has a event loop running.
        """
        if self._invalid:
            self._invalid = False
        wait_time = self._calculate_wait_time()
        if self._task and not self._task.done():
            self._task.cancel()
        self._task = asyncio.create_task(
            self.content_control._filter_choices(wait_time)
        )
        self._task.add_done_callback(self._filter_callback)

    def _handle_toggle_choice(self, _) -> None:
        """Handle tab event, alter the `selected` state of the choice."""
        if not self._multiselect:
            return
        current_selected_index = self.content_control.selection["index"]
        self.content_control.choices[current_selected_index][
            "enabled"
        ] = not self.content_control.choices[current_selected_index]["enabled"]

    def _handle_enter(self, event: "KeyPressEvent") -> None:
        """Handle enter event.

        Validate the result first.

        In multiselect scenario, if no TAB is entered, then capture the current
        highlighted choice and return the value in a list.
        Otherwise, return all TAB choices as a list.

        In normal scenario, reutrn the current highlighted choice.

        If current UI contains no choice due to filter, return None.
        """
        try:
            fake_document = FakeDocument(self.result_value)
            self._validator.validate(fake_document)  # type: ignore
            if self._multiselect:
                self.status["answered"] = True
                if not self.selected_choices:
                    self.status["result"] = [self.content_control.selection["name"]]
                    event.app.exit(result=[self.content_control.selection["value"]])
                else:
                    self.status["result"] = self.result_name
                    event.app.exit(result=self.result_value)
            else:
                self.status["answered"] = True
                self.status["result"] = self.content_control.selection["name"]
                event.app.exit(result=self.content_control.selection["value"])
        except ValidationError as e:
            self._set_error(str(e))
        except IndexError:
            self.status["answered"] = True
            self.status["result"] = None if not self._multiselect else []
            event.app.exit(result=None if not self._multiselect else [])

    @property
    def content_control(self) -> InquirerPyFuzzyControl:
        """InquirerPyFuzzyControl: Override for type-hinting."""
        return cast(InquirerPyFuzzyControl, super().content_control)

    @content_control.setter
    def content_control(self, value: InquirerPyFuzzyControl) -> None:
        self._content_control = value

    def _get_current_text(self) -> str:
        """Get current input buffer text."""
        return self._buffer.text



from InquirerPy import inquirer
from InquirerPy.validator import NumberValidator


def main():
    """Alternate syntax example."""

    name = inquirer.text(message="Enter your name:").execute()
    company = inquirer.text(
        message="Which company would you like to apply:",
        completer={
            "Google": None,
            "Facebook": None,
            "Amazon": None,
            "Netflix": None,
            "Apple": None,
            "Microsoft": None,
        },
        multicolumn_complete=True,
    ).execute()
    salary = inquirer.text(
        message="What's your salary expectation(k):",
        transformer=lambda result: "%sk" % result,
        filter=lambda result: int(result) * 1000,
        validate=NumberValidator(),
    ).execute()


if __name__ == "__main__":
    main()



from InquirerPy import prompt
from InquirerPy.validator import NumberValidator


def main():
    """Classic syntax example."""
    questions = [
        {"type": "input", "message": "Enter your name:"},
        {
            "type": "input",
            "message": "Which company would you like to apply:",
            "completer": {
                "Google": None,
                "Facebook": None,
                "Amazon": None,
                "Netflix": None,
                "Apple": None,
                "Microsoft": None,
            },
            "multicolumn_complete": True,
        },
        {
            "type": "input",
            "message": "What's your salary expectation(k):",
            "transformer": lambda result: "%sk" % result,
            "filter": lambda result: int(result) * 1000,
            "validate": NumberValidator(),
        },
    ]

    result = prompt(questions)


if __name__ == "__main__":
    main()



"""This module contains the classic entrypoint for creating prompts.

A `PyInquirer <https://github.com/CITGuru/PyInquirer>`_ compatible entrypoint :func:`.prompt`.
"""
from typing import Any, Dict, List, Optional, Tuple, Union

from InquirerPy.exceptions import InvalidArgument, RequiredKeyNotFound
from InquirerPy.prompts.checkbox import CheckboxPrompt
from InquirerPy.prompts.confirm import ConfirmPrompt
from InquirerPy.prompts.expand import ExpandPrompt
from InquirerPy.prompts.filepath import FilePathPrompt
from InquirerPy.prompts.fuzzy import FuzzyPrompt
from InquirerPy.prompts.input import InputPrompt
from InquirerPy.prompts.list import ListPrompt
from InquirerPy.prompts.number import NumberPrompt
from InquirerPy.prompts.rawlist import RawlistPrompt
from InquirerPy.prompts.secret import SecretPrompt
from InquirerPy.utils import (
    InquirerPyKeybindings,
    InquirerPyQuestions,
    InquirerPySessionResult,
    get_style,
)

__all__ = ["prompt", "prompt_async"]

question_mapping = {
    "confirm": ConfirmPrompt,
    "filepath": FilePathPrompt,
    "password": SecretPrompt,
    "input": InputPrompt,
    "list": ListPrompt,
    "checkbox": CheckboxPrompt,
    "rawlist": RawlistPrompt,
    "expand": ExpandPrompt,
    "fuzzy": FuzzyPrompt,
    "number": NumberPrompt,
}


def _get_questions(questions: InquirerPyQuestions) -> List[Dict[str, Any]]:
    """Process and validate questions.

    Args:
        questions: List of questions to create prompt.

    Returns:
        List of validated questions.
    """
    if isinstance(questions, dict):
        questions = [questions]

    if not isinstance(questions, list):
        raise InvalidArgument("argument questions should be type of list or dictionary")

    return questions


def _get_question(
    original_question: Dict[str, Any], result: InquirerPySessionResult, index: int
) -> Tuple[Optional[Dict[str, Any]], str, Union[str, int], str]:
    """Get information from individual question.

    Args:
        original_question: Original question dictionary.
        result: Current prompt session result.
        index: Question index.

    Returns:
        A tuple containing question information in the order of
            question dictionary, type of question, name of question, message of question.
    """
    question = original_question.copy()
    question_type = question.pop("type")
    question_name = question.pop("name", index)
    message = question.pop("message")
    question_when = question.pop("when", None)
    if question_when and not question_when(result):
        result[question_name] = None
        question = None
    return question, question_type, question_name, message


async def prompt_async(
    questions: InquirerPyQuestions,
    style: Optional[Dict[str, str]] = None,
    vi_mode: bool = False,
    raise_keyboard_interrupt: bool = True,
    keybindings: Optional[InquirerPyKeybindings] = None,
    style_override: bool = True,
) -> InquirerPySessionResult:
    """Classic syntax entrypoint to create a prompt session via asynchronous method.

    Refer to :func:`InquirerPy.resolver.prompt` for detailed documentations.
    """
    result: InquirerPySessionResult = {}
    if not keybindings:
        keybindings = {}

    questions = _get_questions(questions=questions)
    question_style = get_style(style, style_override)

    for index, original_question in enumerate(questions):
        try:
            question, question_type, question_name, message = _get_question(
                original_question=original_question, result=result, index=index
            )
            if question is None:
                continue
            args = {
                "message": message,
                "style": question_style,
                "vi_mode": vi_mode,
                "raise_keyboard_interrupt": raise_keyboard_interrupt,
                "session_result": result,
                "keybindings": {**keybindings, **question.pop("keybindings", {})},
            }
            result[question_name] = await question_mapping[question_type](
                **args, **question
            ).execute_async()
        except KeyError:
            raise RequiredKeyNotFound

    return result


def prompt(
    questions: InquirerPyQuestions,
    style: Optional[Dict[str, str]] = None,
    vi_mode: bool = False,
    raise_keyboard_interrupt: bool = True,
    keybindings: Optional[InquirerPyKeybindings] = None,
    style_override: bool = True,
) -> InquirerPySessionResult:
    """Classic syntax entrypoint to create a prompt session.

    Resolve user provided list of questions, display prompts and get the results.

    Args:
        questions: A list of :ref:`pages/prompt:question` to ask. Refer to documentation for more info.
        style: A :class:`dict` containing the style specification for the prompt. Refer to :ref:`pages/style:Style` for more info.
        vi_mode: Use vim keybindings for the prompt instead of the default emacs keybindings.
            Refer to :ref:`pages/kb:Keybindings` for more info.
        raise_keyboard_interrupt: Raise the :class:`KeyboardInterrupt` exception when `ctrl-c` is pressed. If false, the result
            will be `None` and the question is skiped.
        keybindings: List of custom :ref:`pages/kb:Keybindings` to apply. Refer to documentation for more info.
        style_override: Override all default styles. When providing any style customisation, all default styles are removed when this is True.

    Returns:
        A dictionary containing all of the question answers. The key is the name of the question and the value is the
        user answer. If the `name` key is not present as part of the question, then the question index will be used
        as the key.

    Raises:
        RequiredKeyNotFound: When the question is missing required keys.
        InvalidArgument: When the provided `questions` argument is not a type of :class:`list` nor :class:`dictionary`.

    Examples:
        >>> from InquirerPy import prompt
        >>> from InquirerPy.validator import NumberValidator
        >>> questions = [
        ...     {
        ...         "type": "input",
        ...         "message": "Enter your age:",
        ...         "validate": NumberValidator(),
        ...         "invalid_message": "Input should be number.",
        ...         "default": "18",
        ...         "name": "age",
        ...         "filter": lambda result: int(result),
        ...         "transformer": lambda result: "Adult" if int(result) >= 18 else "Youth",
        ...     },
        ...     {
        ...         "type": "rawlist",
        ...         "message": "What drinks would you like to buy:",
        ...         "default": 2,
        ...         "choices": lambda result: ["Soda", "Cidr", "Water", "Milk"]
        ...         if result["age"] < 18
        ...         else ["Wine", "Beer"],
        ...         "name": "drink",
        ...     },
        ...     {
        ...         "type": "list",
        ...         "message": "Would you like a bag:",
        ...         "choices": ["Yes", "No"],
        ...         "when": lambda result: result["drink"] in {"Wine", "Beer"},
        ...     },
        ...     {"type": "confirm", "message": "Confirm?", "default": True},
        ... ]
        >>> result = prompt(questions=questions)
    """
    result: InquirerPySessionResult = {}
    if not keybindings:
        keybindings = {}

    questions = _get_questions(questions=questions)
    question_style = get_style(style, style_override)

    for index, original_question in enumerate(questions):
        try:
            question, question_type, question_name, message = _get_question(
                original_question=original_question, result=result, index=index
            )
            if question is None:
                continue
            args = {
                "message": message,
                "style": question_style,
                "vi_mode": vi_mode,
                "raise_keyboard_interrupt": raise_keyboard_interrupt,
                "session_result": result,
                "keybindings": {**keybindings, **question.pop("keybindings", {})},
            }
            result[question_name] = question_mapping[question_type](
                **args, **question
            ).execute()
        except KeyError:
            raise RequiredKeyNotFound

    return result



"""Module contains :class:`.InstructionWindow` which can be used to display long instructions."""

from typing import TYPE_CHECKING

from prompt_toolkit.layout.containers import ConditionalContainer, Window
from prompt_toolkit.layout.controls import FormattedTextControl

if TYPE_CHECKING:
    from prompt_toolkit.filters.base import FilterOrBool
    from prompt_toolkit.formatted_text.base import AnyFormattedText


class InstructionWindow(ConditionalContainer):
    """Conditional `prompt_toolkit` :class:`~prompt_toolkit.layout.Window` that displays long instructions.

    Args:
        message: Long instructions to display.
        filter: Condition to display the instruction window.
    """

    def __init__(self, message: str, filter: "FilterOrBool", **kwargs) -> None:
        self._message = message
        super().__init__(
            Window(
                FormattedTextControl(text=self._get_message),
                dont_extend_height=True,
                **kwargs
            ),
            filter=filter,
        )

    def _get_message(self) -> "AnyFormattedText":
        """Get long instruction to display.

        Returns:
            FormattedText in list of tuple format.
        """
        return [("class:long_instruction", self._message)]



"""Module contains base class for prompts.

BaseSimplePrompt ← InputPrompt ← SecretPrompt ...
        ↑
BaseComplexPrompt
        ↑
BaseListPrompt ← FuzzyPrompt
        ↑
ListPrompt ← ExpandPrompt ...
"""

from .complex import BaseComplexPrompt, FakeDocument
from .control import Choice, InquirerPyUIListControl
from .list import BaseListPrompt
from .simple import BaseSimplePrompt



from InquirerPy import prompt


def main():
    questions = [
        {
            "type": "confirm",
            "message": "Proceed?",
            "name": "proceed",
            "default": True,
        },
        {
            "type": "confirm",
            "message": "Require 1 on 1?",
            "when": lambda result: result["proceed"],
        },
        {
            "type": "confirm",
            "message": "Confirm?",
            "when": lambda result: result.get("1", False),
        },
    ]

    result = prompt(questions)


if __name__ == "__main__":
    main()



# NOTE: Following example requires boto3 package.
import os

import boto3

from InquirerPy import inquirer
from InquirerPy.exceptions import InvalidArgument
from InquirerPy.validator import PathValidator

client = boto3.client("s3")
os.environ["INQUIRERPY_VI_MODE"] = "true"


def get_bucket(_):
    return [bucket["Name"] for bucket in client.list_buckets()["Buckets"]]


def walk_s3_bucket(bucket):
    response = []
    paginator = client.get_paginator("list_objects")
    for result in paginator.paginate(Bucket=bucket):
        for file in result["Contents"]:
            response.append(file["Key"])
    return response


try:
    action = inquirer.select(
        message="Select an S3 action:", choices=["Upload", "Download"]
    ).execute()

    if action == "Upload":
        file_to_upload = inquirer.filepath(
            message="Enter the filepath to upload:",
            validate=PathValidator(),
            only_files=True,
        ).execute()
        bucket = inquirer.fuzzy(
            message="Select a bucket:", choices=get_bucket, spinner_enable=True
        ).execute()
    else:
        bucket = inquirer.fuzzy(
            message="Select a bucket:", choices=get_bucket, spinner_enable=True
        ).execute()
        file_to_download = inquirer.fuzzy(
            message="Select files to download:",
            choices=lambda _: walk_s3_bucket(bucket),
            multiselect=True,
            spinner_enable=True,
        ).execute()
        destination = inquirer.filepath(
            message="Enter destination folder:",
            only_directories=True,
            validate=PathValidator(),
        ).execute()

    confirm = inquirer.confirm(message="Confirm?").execute()
except InvalidArgument:
    print("No available choices")

# Download or Upload the file based on result ...



from InquirerPy import prompt
from InquirerPy.enum import INQUIRERPY_POINTER_SEQUENCE
from InquirerPy.validator import EmptyInputValidator

questions = [
    {
        "message": "Delivery or Takeaway?",
        "type": "list",
        "choices": ["Takeaway", "Delivery"],
    },
    {
        "message": "What's your name?",
        "type": "input",
        "validate": EmptyInputValidator(),
    },
    {
        "message": "What's your address",
        "type": "input",
        "validate": EmptyInputValidator("Address cannot be empty"),
        "when": lambda x: x[0] == "Delivery",
    },
    {
        "message": "What pizza would you like?",
        "type": "rawlist",
        "choices": [
            "Pepperoni",
            "Hawaii",
            "Simple Cheese",
            "Peri Peri Chicken",
            "Meath Lover",
        ],
        "pointer": INQUIRERPY_POINTER_SEQUENCE,
    },
    {
        "message": "Select toppings:",
        "type": "fuzzy",
        "choices": [
            "Pepperoni",
            "Mushrooms",
            "Sausage",
            "Onions",
            "Bacon",
            "Extra Cheese",
            "Peppers",
            "Black Olives",
            "Chicken",
            "Pineapple",
            "Spinach",
            "Fresh Basil",
            "Ham",
            "Pesto",
            "Beef",
        ],
        "multiselect": True,
    },
    {"message": "Confirm order?", "type": "confirm", "default": False},
]
result = prompt(
    questions,
    style={"questionmark": "#ff9d00 bold"},
    vi_mode=True,
    style_override=False,
)



from InquirerPy import inquirer
from InquirerPy.validator import EmptyInputValidator


def main() -> None:
    integer_val = inquirer.number(
        message="Enter integer:",
        min_allowed=-2,
        max_allowed=10,
        validate=EmptyInputValidator(),
    ).execute()
    float_val = inquirer.number(
        message="Enter float:",
        float_allowed=True,
        validate=EmptyInputValidator(),
    ).execute()


if __name__ == "__main__":
    main()



import unittest
from decimal import Decimal
from unittest.mock import ANY, call, patch

from prompt_toolkit.keys import Keys

from InquirerPy.exceptions import InvalidArgument
from InquirerPy.prompts.number import NumberPrompt


class TestNumberPrompt(unittest.TestCase):
    def setUp(self) -> None:
        self.prompt = NumberPrompt(
            message="Hello",
            default=1,
            max_allowed=10,
            min_allowed=-2,
            replace_mode=True,
        )
        self.float_prompt = NumberPrompt(
            message="Hello",
            float_allowed=True,
            default=1,
            max_allowed=10,
            min_allowed=-2,
            replace_mode=True,
        )
        self.prompt._on_rendered(None)
        self.float_prompt._on_rendered(None)

    def test_contructor(self) -> None:
        self.assertFalse(self.prompt._float)
        self.assertEqual(self.prompt._default, 1)
        self.assertFalse(self.prompt._is_float())
        self.assertEqual(self.prompt.focus, self.prompt._whole_window)

        prompt = NumberPrompt(message="", default=lambda _: 1)
        self.assertEqual(prompt._default, 1)

        try:
            NumberPrompt(message="", default="asdfasd")
            NumberPrompt(message="", default="asdfas", float_allowed=True)
        except InvalidArgument:
            pass
        else:
            self.fail("InvalidArgument should be raised")

    def test_float_constructor(self) -> None:
        self.assertTrue(self.float_prompt._float)
        self.assertEqual(self.float_prompt._default, 1.0)
        self.assertTrue(self.float_prompt._is_float())

    @patch("InquirerPy.prompts.number.NumberPrompt.register_kb")
    def test_kb_registered(self, mocked_kb) -> None:
        prompt = NumberPrompt(message="")
        mocked_kb.assert_has_calls([call(Keys.Any)])

        prompt._after_render(None)
        mocked_kb.assert_has_calls([call("down", filter=ANY)])
        mocked_kb.assert_has_calls([call("up", filter=ANY)])
        mocked_kb.assert_has_calls([call("left", filter=ANY)])
        mocked_kb.assert_has_calls([call(Keys.Tab, filter=ANY)])
        for i in range(10):
            mocked_kb.assert_has_calls([call(str(i), filter=ANY)])

    def test_on_rendered(self) -> None:
        self.prompt._on_rendered(None)
        self.assertEqual(self.prompt._whole_buffer.text, "1")
        self.assertEqual(self.prompt._integral_buffer.text, "0")
        self.assertFalse(self.prompt._whole_replace)
        self.assertEqual(self.prompt._whole_buffer.cursor_position, 1)

        self.float_prompt._default = 1.1
        self.float_prompt._on_rendered(None)
        self.assertEqual(self.float_prompt._whole_buffer.text, "1")
        self.assertEqual(self.float_prompt._integral_buffer.text, "1")
        self.assertTrue(self.prompt._integral_buffer)
        self.assertEqual(self.prompt._integral_buffer.cursor_position, 0)

    def test_on_rendered_sn(self) -> None:
        prompt = NumberPrompt(
            message="", default=0.000000123, float_allowed=True, replace_mode=True
        )
        prompt._on_rendered(None)
        self.assertTrue(prompt._whole_replace)
        self.assertFalse(prompt._integral_replace)
        self.assertEqual(prompt._whole_buffer.text, "0")
        self.assertEqual(prompt._integral_buffer.text, "000000123")

        prompt = NumberPrompt(
            message="", default=0.00000000099912312, float_allowed=True
        )
        prompt._on_rendered(None)
        self.assertEqual(prompt._integral_buffer.text, "00000000099912312")

    def test_handle_down(self) -> None:
        self.prompt._on_rendered(None)
        self.prompt._handle_down(None)
        self.assertEqual(self.prompt._whole_buffer.text, "0")
        self.prompt._handle_down(None)
        self.assertEqual(self.prompt._whole_buffer.text, "-1")
        self.prompt._handle_down(None)
        self.prompt._handle_down(None)
        self.prompt._handle_down(None)
        self.assertEqual(self.prompt._whole_buffer.text, "-2")

        self.prompt._whole_buffer.text = ""
        self.prompt._handle_down(None)
        self.assertEqual(self.prompt._whole_buffer.text, "0")

    def test_handle_down_float(self) -> None:
        self.float_prompt._default = 0.3
        self.float_prompt._on_rendered(None)
        self.float_prompt._handle_focus(None)
        self.float_prompt._handle_down(None)
        self.assertEqual(self.float_prompt._integral_buffer.text, "2")
        self.float_prompt._handle_down(None)
        self.float_prompt._handle_down(None)
        self.float_prompt._handle_down(None)
        self.float_prompt._handle_down(None)
        self.assertEqual(self.float_prompt._integral_buffer.text, "0")

    def test_handle_up(self) -> None:
        self.prompt._on_rendered(None)
        self.prompt._handle_up(None)
        self.assertEqual(self.prompt._whole_buffer.text, "2")
        self.prompt._handle_up(None)
        self.prompt._handle_up(None)
        self.prompt._handle_up(None)
        self.assertEqual(self.prompt._whole_buffer.text, "5")
        self.prompt._handle_up(None)
        self.prompt._handle_up(None)
        self.prompt._handle_up(None)
        self.prompt._handle_up(None)
        self.prompt._handle_up(None)
        self.prompt._handle_up(None)
        self.prompt._handle_up(None)
        self.assertEqual(self.prompt._whole_buffer.text, "10")

        self.prompt._whole_buffer.text = ""
        self.prompt._handle_up(None)
        self.assertEqual(self.prompt._whole_buffer.text, "0")

    def test_handle_up_float(self) -> None:
        self.float_prompt._default = 9.0
        self.float_prompt._on_rendered(None)
        self.float_prompt._handle_focus(None)
        self.float_prompt._handle_up(None)
        self.assertEqual(self.float_prompt._integral_buffer.text, "1")
        self.float_prompt._handle_up(None)
        self.assertEqual(self.float_prompt._integral_buffer.text, "2")
        self.float_prompt._handle_focus(None)
        self.float_prompt._handle_up(None)
        self.assertEqual(self.float_prompt._integral_buffer.text, "2")

    def test_handle_left(self) -> None:
        self.prompt._on_rendered(None)
        self.assertEqual(self.prompt._whole_buffer.cursor_position, 1)
        self.prompt._handle_left(None)
        self.assertEqual(self.prompt._whole_buffer.cursor_position, 0)
        self.prompt._handle_left(None)
        self.assertEqual(self.prompt._whole_buffer.cursor_position, 0)

    def test_handle_left_float(self) -> None:
        self.float_prompt._on_rendered(None)
        self.assertEqual(self.float_prompt._whole_buffer.cursor_position, 1)
        self.assertEqual(self.float_prompt._integral_buffer.cursor_position, 0)
        self.float_prompt._integral_buffer.cursor_position = 1
        self.float_prompt._handle_focus(None)
        self.float_prompt._handle_left(None)
        self.assertEqual(self.float_prompt._integral_buffer.cursor_position, 0)
        self.assertEqual(self.float_prompt.focus, self.float_prompt._integral_window)
        self.float_prompt._handle_left(None)
        self.assertEqual(self.float_prompt._integral_buffer.cursor_position, 0)
        self.assertEqual(self.float_prompt.focus, self.float_prompt._whole_window)
        self.assertEqual(self.float_prompt._whole_buffer.cursor_position, 1)
        self.float_prompt._handle_left(None)
        self.assertEqual(self.float_prompt._whole_buffer.cursor_position, 0)

    def test_handle_right(self) -> None:
        self.prompt._on_rendered(None)
        self.assertEqual(self.prompt._whole_buffer.cursor_position, 1)
        self.assertEqual(self.prompt.focus, self.prompt._whole_window)
        self.prompt._handle_right(None)
        self.assertEqual(self.prompt._whole_buffer.cursor_position, 1)
        self.assertNotEqual(self.prompt.focus, self.prompt._integral_window)

        self.prompt._whole_buffer.cursor_position = 0
        self.assertEqual(self.prompt._whole_buffer.cursor_position, 0)
        self.prompt._handle_right(None)
        self.assertEqual(self.prompt._whole_buffer.cursor_position, 1)

    def test_handle_right_float(self) -> None:
        self.float_prompt._on_rendered(None)
        self.assertEqual(self.float_prompt._whole_buffer.cursor_position, 1)
        self.assertEqual(self.float_prompt.focus, self.float_prompt._whole_window)
        self.float_prompt._handle_right(None)
        self.assertEqual(self.float_prompt.focus, self.float_prompt._integral_window)
        self.assertEqual(self.float_prompt._integral_buffer.cursor_position, 0)
        self.float_prompt._handle_right(None)
        self.assertEqual(self.float_prompt._integral_buffer.cursor_position, 1)
        self.float_prompt._handle_right(None)
        self.assertEqual(self.float_prompt._integral_buffer.cursor_position, 1)

    def test_handle_enter(self) -> None:
        self.prompt._on_rendered(None)
        with patch("prompt_toolkit.utils.Event") as mock:
            event = mock.return_value
            self.prompt._handle_enter(event)
        self.assertTrue(self.prompt.status["answered"])
        self.assertEqual(self.prompt.status["result"], "1")

        self.prompt._whole_buffer.text = ""
        with patch("prompt_toolkit.utils.Event") as mock:
            event = mock.return_value
            self.prompt._handle_enter(event)
        self.assertTrue(self.prompt.status["answered"])
        self.assertEqual(self.prompt.status["result"], "")

    def test_handle_enter_float(self) -> None:
        self.float_prompt._on_rendered(None)
        with patch("prompt_toolkit.utils.Event") as mock:
            event = mock.return_value
            self.float_prompt._handle_enter(event)
        self.assertTrue(self.float_prompt.status["answered"])
        self.assertEqual(self.float_prompt.status["result"], "1.0")

        self.float_prompt._integral_buffer.text = ""
        with patch("prompt_toolkit.utils.Event") as mock:
            event = mock.return_value
            self.float_prompt._handle_enter(event)
        self.assertTrue(self.float_prompt.status["answered"])
        self.assertEqual(self.float_prompt.status["result"], "1.0")

        self.float_prompt._integral_buffer.text = ""
        self.float_prompt._whole_buffer.text = ""
        with patch("prompt_toolkit.utils.Event") as mock:
            event = mock.return_value
            self.float_prompt._handle_enter(event)
        self.assertTrue(self.float_prompt.status["answered"])
        self.assertEqual(self.float_prompt.status["result"], "")

    def test_handle_enter_validation(self) -> None:
        prompt = NumberPrompt(message="", validate=lambda x: x == 1)
        prompt._on_rendered(None)
        with patch("prompt_toolkit.utils.Event") as mock:
            event = mock.return_value
            prompt._handle_enter(event)
        self.assertFalse(prompt.status["answered"])
        self.assertEqual(prompt.status["result"], None)
        self.assertEqual(
            prompt._get_error_message(), [("class:validation-toolbar", "Invalid input")]
        )

    def test_handle_focus(self) -> None:
        self.assertEqual(self.prompt.focus, self.prompt._whole_window)
        self.prompt._handle_focus(None)
        self.assertEqual(self.prompt.focus, self.prompt._whole_window)

    def test_handle_focus_float(self) -> None:
        self.assertEqual(self.float_prompt.focus, self.float_prompt._whole_window)
        self.float_prompt._handle_focus(None)
        self.assertEqual(self.float_prompt.focus, self.float_prompt._integral_window)
        self.float_prompt._handle_focus(None)
        self.assertEqual(self.float_prompt.focus, self.float_prompt._whole_window)

    def test_handle_input(self) -> None:
        with patch("prompt_toolkit.utils.Event") as mock:
            event = mock.return_value
            self.prompt._whole_buffer.cursor_position = 0
            self.prompt._handle_input(event)

    @patch("InquirerPy.prompts.number.NumberPrompt._on_text_change")
    def test_on_text_change(self, mocked_text) -> None:
        self.prompt._whole_buffer.text = "10"
        mocked_text.assert_called()
        self.assertEqual(self.prompt._whole_width, 3)

        mocked_text.reset_mock()
        self.float_prompt._integral_buffer.text = "10"
        mocked_text.assert_called()
        self.assertEqual(self.float_prompt._integral_width, 3)
        self.float_prompt._integral_buffer.text = "100"
        self.assertEqual(self.float_prompt._integral_width, 4)

    def test_handle_negative_toggle(self) -> None:
        self.assertEqual(self.prompt._whole_buffer.text, "1")
        self.assertEqual(self.prompt._whole_buffer.cursor_position, 1)
        self.prompt._handle_negative_toggle(None)
        self.assertEqual(self.prompt._whole_buffer.text, "-1")
        self.assertEqual(self.prompt._whole_buffer.cursor_position, 2)
        self.prompt._handle_negative_toggle(None)
        self.assertEqual(self.prompt._whole_buffer.text, "1")
        self.assertEqual(self.prompt._whole_buffer.cursor_position, 1)

        self.prompt._min = -10
        self.prompt._max = 10
        self.prompt._whole_buffer.text = "10"

        self.prompt._whole_buffer.cursor_position = 2
        self.prompt._handle_negative_toggle(None)
        self.assertEqual(self.prompt._whole_buffer.text, "-10")
        self.assertEqual(self.prompt._whole_buffer.cursor_position, 3)
        self.prompt._handle_negative_toggle(None)
        self.assertEqual(self.prompt._whole_buffer.text, "10")
        self.assertEqual(self.prompt._whole_buffer.cursor_position, 2)

        self.prompt._whole_buffer.cursor_position = 1
        self.prompt._handle_negative_toggle(None)
        self.assertEqual(self.prompt._whole_buffer.text, "-10")
        self.assertEqual(self.prompt._whole_buffer.cursor_position, 2)
        self.prompt._handle_negative_toggle(None)
        self.assertEqual(self.prompt._whole_buffer.text, "10")
        self.assertEqual(self.prompt._whole_buffer.cursor_position, 1)

        self.prompt._whole_buffer.cursor_position = 0
        self.prompt._handle_negative_toggle(None)
        self.assertEqual(self.prompt._whole_buffer.text, "-10")
        self.assertEqual(self.prompt._whole_buffer.cursor_position, 1)
        self.prompt._handle_negative_toggle(None)
        self.assertEqual(self.prompt._whole_buffer.text, "10")
        self.assertEqual(self.prompt._whole_buffer.cursor_position, 0)

        self.prompt._whole_buffer.text = "-"
        self.prompt._handle_negative_toggle(None)
        self.assertEqual(self.prompt._whole_buffer.text, "0")

    def test_cursor_position(self) -> None:
        self.prompt._handle_negative_toggle(None)
        self.prompt._whole_buffer.cursor_position = 0
        self.assertEqual(self.prompt._whole_buffer.cursor_position, 1)

    def test_value(self) -> None:
        self.prompt._whole_buffer.text = "asdfad"
        self.assertEqual(self.prompt.value, 1)
        self.assertEqual(
            self.prompt._get_error_message(),
            [("class:validation-toolbar", "Remove any non-integer value")],
        )

    def test_fix_sn(self) -> None:
        left, right = self.prompt._fix_sn("0E-7")
        self.assertEqual(left, "0")
        self.assertEqual(right, "0000000")

        left, right = self.prompt._fix_sn("1.2E-7")
        self.assertEqual(left, "0")
        self.assertEqual(right, "00000012")

        left, right = self.prompt._fix_sn("9.88E-11")
        self.assertEqual(left, "0")
        self.assertEqual(right, "0000000000988")

    def test_handle_number(self) -> None:
        self.prompt._whole_buffer.text = "0"
        self.prompt._handle_number(increment=True)
        self.assertEqual(self.prompt._whole_buffer.text, "1")
        self.assertEqual(self.prompt._whole_buffer.cursor_position, 1)

        self.prompt._whole_buffer.text = ""
        self.prompt._handle_number(increment=True)
        self.assertEqual(self.prompt._whole_buffer.text, "0")
        self.assertEqual(self.prompt._whole_buffer.cursor_position, 1)
        self.prompt._handle_number(increment=True)
        self.assertEqual(self.prompt._whole_buffer.text, "1")
        self.assertEqual(self.prompt._whole_buffer.cursor_position, 1)

        self.prompt._whole_buffer.text = "9"
        self.prompt._handle_number(increment=True)
        self.assertEqual(self.prompt._whole_buffer.text, "10")
        self.assertEqual(self.prompt._whole_buffer.cursor_position, 2)
        self.prompt._handle_number(increment=False)
        self.assertEqual(self.prompt._whole_buffer.text, "9")
        self.assertEqual(self.prompt._whole_buffer.cursor_position, 1)
        self.prompt._whole_buffer.cursor_position = 0
        self.prompt._handle_number(increment=True)
        self.assertEqual(self.prompt._whole_buffer.text, "10")
        self.assertEqual(self.prompt._whole_buffer.cursor_position, 1)
        self.prompt._handle_number(increment=False)
        self.assertEqual(self.prompt._whole_buffer.text, "9")
        self.assertEqual(self.prompt._whole_buffer.cursor_position, 0)

        self.prompt._whole_buffer.text = "0"
        self.prompt._whole_buffer.cursor_position = 0
        self.prompt._handle_number(increment=False)
        self.assertEqual(self.prompt._whole_buffer.text, "-1")
        self.assertEqual(self.prompt._whole_buffer.cursor_position, 1)
        self.prompt._handle_number(increment=True)
        self.assertEqual(self.prompt._whole_buffer.text, "0")
        self.assertEqual(self.prompt._whole_buffer.cursor_position, 0)

    def test_handle_number_float(self) -> None:
        self.assertEqual(self.float_prompt._integral_buffer.text, "0")
        self.assertEqual(self.float_prompt._integral_buffer.cursor_position, 0)
        self.assertTrue(self.float_prompt._integral_replace)
        self.float_prompt._handle_focus(None)
        self.float_prompt._handle_number(increment=True)
        self.assertEqual(self.float_prompt._integral_buffer.text, "1")
        # increased due to buffer starts as replace mode
        self.assertFalse(self.float_prompt._integral_replace)
        self.assertEqual(self.float_prompt._integral_buffer.cursor_position, 1)

        self.float_prompt._integral_buffer.cursor_position = 0
        self.float_prompt._integral_buffer.text = "001"
        self.float_prompt._handle_number(increment=True)
        self.assertEqual(self.float_prompt._integral_buffer.text, "002")
        self.assertEqual(self.float_prompt._integral_buffer.cursor_position, 0)
        self.float_prompt._handle_number(increment=True)
        self.assertEqual(self.float_prompt._integral_buffer.text, "003")
        self.assertEqual(self.float_prompt._integral_buffer.cursor_position, 0)
        self.float_prompt._handle_number(increment=False)
        self.assertEqual(self.float_prompt._integral_buffer.text, "002")
        self.assertEqual(self.float_prompt._integral_buffer.cursor_position, 0)
        self.float_prompt._integral_buffer.text = "009"
        self.float_prompt._handle_number(increment=True)
        self.assertEqual(self.float_prompt._integral_buffer.text, "0010")
        self.assertEqual(self.float_prompt._integral_buffer.cursor_position, 1)
        self.float_prompt._handle_number(increment=False)
        self.assertEqual(self.float_prompt._integral_buffer.text, "009")
        self.assertEqual(self.float_prompt._integral_buffer.cursor_position, 0)

    def test_handle_dot(self) -> None:
        self.assertEqual(self.prompt.focus_buffer, self.prompt._whole_buffer)
        self.prompt._handle_dot(None)
        self.assertEqual(self.prompt.focus_buffer, self.prompt._whole_buffer)

        self.assertEqual(
            self.float_prompt.focus_buffer, self.float_prompt._whole_buffer
        )
        self.float_prompt._handle_dot(None)
        self.assertEqual(
            self.float_prompt.focus_buffer, self.float_prompt._integral_buffer
        )
        self.float_prompt._handle_dot(None)
        self.assertEqual(
            self.float_prompt.focus_buffer, self.float_prompt._integral_buffer
        )

    def test_sn(self) -> None:
        self.float_prompt.value = Decimal("0.00000000099912312")
        self.assertEqual(self.float_prompt._integral_buffer.text, "00000000099912312")

        self.float_prompt.value = Decimal("1.1")
        self.assertEqual(self.float_prompt._integral_buffer.text, "1")

    def test_replace_mode(self) -> None:
        self.assertFalse(self.prompt._whole_replace)
        self.assertTrue(self.float_prompt._integral_replace)
        self.float_prompt._handle_focus(None)
        self.float_prompt._handle_left(None)
        self.assertFalse(self.float_prompt._integral_replace)

        prompt = NumberPrompt(message="", float_allowed=True, replace_mode=True)
        prompt._on_rendered(None)
        self.assertTrue(prompt._whole_replace)
        self.assertTrue(prompt._integral_replace)
        prompt._handle_up(None)
        self.assertFalse(prompt._whole_replace)
        prompt._handle_dot(None)
        prompt._handle_right(None)
        self.assertFalse(prompt._integral_replace)

    def test_replace_mode_no(self) -> None:
        prompt = NumberPrompt(message="", replace_mode=False)
        prompt._on_rendered(None)
        self.assertFalse(prompt._whole_replace)
        self.assertFalse(prompt._integral_replace)

    def test_empty_default(self) -> None:
        prompt = NumberPrompt(message="", default=None)
        prompt._on_rendered(None)
        self.assertEqual(prompt._whole_buffer.text, "")
        self.assertEqual(prompt._integral_buffer.text, "")

        prompt = NumberPrompt(message="", default=None, float_allowed=True)
        prompt._on_rendered(None)
        self.assertEqual(prompt._whole_buffer.text, "")
        self.assertEqual(prompt._integral_buffer.text, "")

    def test_empty_buffer(self) -> None:
        prompt = NumberPrompt(message="", default=None, float_allowed=True)
        prompt._on_rendered(None)
        self.assertEqual(prompt._whole_buffer.text, "")
        self.assertEqual(prompt._integral_buffer.text, "")
        prompt._whole_buffer.text = "1"
        self.assertEqual(prompt._whole_buffer.text, "1")
        self.assertEqual(prompt._integral_buffer.text, "")
        self.assertEqual(str(prompt.value), "1.0")

        prompt._whole_buffer.text = ""
        prompt._integral_buffer.text = "1"
        self.assertEqual(prompt._whole_buffer.text, "")
        self.assertEqual(prompt._integral_buffer.text, "1")
        self.assertEqual(str(prompt.value), "0.1")



import unittest
from unittest.mock import ANY, call, patch

from prompt_toolkit.buffer import Buffer
from prompt_toolkit.enums import EditingMode
from prompt_toolkit.input import create_pipe_input
from prompt_toolkit.output import DummyOutput
from prompt_toolkit.shortcuts.prompt import CompleteStyle

from InquirerPy.exceptions import InvalidArgument
from InquirerPy.prompts.secret import SecretPrompt
from InquirerPy.utils import InquirerPyStyle
from InquirerPy.validator import PasswordValidator
from tests.style import get_sample_style


class TestSecret(unittest.TestCase):
    def setUp(self):
        self.inp = create_pipe_input()

    def tearDown(self):
        self.inp.close()

    def test_prompt_result(self):
        self.inp.send_text("what\n")
        secret_prompt = SecretPrompt(
            message="hello",
            style=InquirerPyStyle({"answer": ""}),
            default="yes",
            qmark="~",
            vi_mode=False,
            input=self.inp,
            output=DummyOutput(),
        )
        result = secret_prompt.execute()
        self.assertEqual(result, "yeswhat")
        self.assertEqual(
            secret_prompt.status,
            {"answered": True, "result": "yeswhat", "skipped": False},
        )

    @patch.object(Buffer, "validate_and_handle")
    def test_prompt_validation(self, mocked_validate):
        def _hello():
            secret_prompt._session.app.exit(result="yes")

        mocked_validate.side_effect = _hello
        self.inp.send_text("afas\n")
        secret_prompt = SecretPrompt(
            message="what",
            style=InquirerPyStyle({}),
            validate=PasswordValidator(length=8),
            input=self.inp,
            output=DummyOutput(),
        )
        result = secret_prompt.execute()
        mocked_validate.assert_called_once()
        self.assertEqual(result, "yes")
        self.assertEqual(secret_prompt.status["answered"], False)
        self.assertEqual(secret_prompt.status["result"], None)

    def test_prompt_message(self):
        secret_prompt = SecretPrompt(
            message="fooboo", style=InquirerPyStyle({}), qmark="[?]", vi_mode=True
        )
        message = secret_prompt._get_prompt_message()
        self.assertEqual(
            message,
            [
                ("class:questionmark", "[?]"),
                ("class:question", " fooboo"),
                ("class:instruction", " "),
            ],
        )

        secret_prompt.status["answered"] = True
        secret_prompt.status["result"] = "hello"
        message = secret_prompt._get_prompt_message()
        self.assertEqual(
            message,
            [
                ("class:answermark", "?"),
                ("class:answered_question", " fooboo"),
                ("class:answer", " *****"),
            ],
        )

        # instruction
        secret_prompt = SecretPrompt(
            message="fooboo",
            style=InquirerPyStyle({}),
            qmark="[?]",
            vi_mode=True,
            instruction="(abc)",
        )
        message = secret_prompt._get_prompt_message()
        self.assertEqual(
            message,
            [
                ("class:questionmark", "[?]"),
                ("class:question", " fooboo"),
                ("class:instruction", " (abc) "),
            ],
        )

    @patch("InquirerPy.prompts.input.SimpleLexer")
    @patch("InquirerPy.prompts.secret.SecretPrompt._get_prompt_message")
    @patch("InquirerPy.base.simple.Style.from_dict")
    @patch("InquirerPy.base.simple.KeyBindings")
    @patch("InquirerPy.prompts.input.PromptSession")
    def test_callable_called(
        self,
        MockedSession,
        MockedKeyBindings,
        MockedStyle,
        mocked_message,
        MockedLexer,
    ):
        kb = MockedKeyBindings()
        style = MockedStyle()
        lexer = MockedLexer()
        SecretPrompt(
            message="what",
            style=None,
            default="111",
            qmark="[!]",
            vi_mode=True,
        )

        MockedSession.assert_called_once_with(
            message=mocked_message,
            key_bindings=kb,
            style=style,
            completer=None,
            validator=ANY,
            validate_while_typing=False,
            input=None,
            output=None,
            editing_mode=EditingMode.VI,
            lexer=lexer,
            is_password=True,
            multiline=False,
            complete_style=CompleteStyle.COLUMN,
            wrap_lines=True,
            bottom_toolbar=None,
        )
        MockedStyle.assert_has_calls(
            [
                call(),
                call(get_sample_style()),
            ]
        )
        MockedLexer.assert_has_calls([call("class:input")])

    def test_invalid_argument(self):
        self.assertRaises(InvalidArgument, SecretPrompt, "hello", None, 12)
        SecretPrompt(message="hello", default=lambda _: "12")



import asyncio
import unittest
from unittest.mock import ANY, call, patch

from prompt_toolkit.completion.base import CompleteEvent
from prompt_toolkit.document import Document
from prompt_toolkit.enums import EditingMode
from prompt_toolkit.input import create_pipe_input
from prompt_toolkit.output import DummyOutput
from prompt_toolkit.shortcuts.prompt import CompleteStyle

from InquirerPy.enum import INQUIRERPY_POINTER_SEQUENCE
from InquirerPy.prompts.input import InputPrompt


class TestInputPrompt(unittest.TestCase):
    def setUp(self):
        self.inp = create_pipe_input()

    def tearDown(self):
        self.inp.close()

    def test_prompt_result(self):
        self.inp.send_text("hello\n")
        input_prompt = InputPrompt(
            message="yes",
            style=None,
            default="world",
            qmark="!",
            vi_mode=False,
            input=self.inp,
            output=DummyOutput(),
        )
        result = input_prompt.execute()
        self.assertEqual(result, "worldhello")
        self.assertEqual(input_prompt.status["answered"], True)
        self.assertEqual(input_prompt.status["result"], "worldhello")
        self.assertEqual(input_prompt.status["skipped"], False)

    def test_prompt_filter(self):
        self.inp.send_text("hello\n")
        input_prompt = InputPrompt(
            message="yes",
            style=None,
            default="world",
            qmark="!",
            vi_mode=False,
            input=self.inp,
            output=DummyOutput(),
            filter=lambda x: x * 2,
            transformer=lambda _: "what",
        )
        result = input_prompt.execute()
        self.assertEqual(result, "worldhelloworldhello")
        self.assertEqual(input_prompt.status["answered"], True)
        self.assertEqual(input_prompt.status["result"], "worldhello")
        self.assertEqual(
            input_prompt._get_prompt_message(),
            [
                ("class:answermark", "?"),
                ("class:answered_question", " yes"),
                ("class:answer", " what"),
            ],
        )

    def test_prompt_result_multiline(self):
        self.inp.send_text("hello\nworld\nfoo\nboo\x1b\r")
        input_prompt = InputPrompt(
            message="yes",
            style=None,
            default="",
            qmark="!",
            vi_mode=False,
            input=self.inp,
            output=DummyOutput(),
            multiline=True,
        )
        result = input_prompt.execute()
        self.assertEqual(result, "hello\nworld\nfoo\nboo")
        self.assertEqual(input_prompt.status["answered"], True)
        self.assertEqual(input_prompt.status["result"], "hello\nworld\nfoo\nboo")
        self.assertEqual(input_prompt.status["skipped"], False)

    def test_prompt_completion(self):
        input_prompt = InputPrompt(
            message="yes",
            style=None,
            default="",
            qmark="!",
            vi_mode=False,
            input=self.inp,
            output=DummyOutput(),
            multiline=True,
            completer={"hello": None, "hey": None, "what": None},
        )

        completer = input_prompt._completer
        doc_text = "he"
        doc = Document(doc_text, len(doc_text))
        event = CompleteEvent()
        completions = [
            completion.text
            for completion in list(completer.get_completions(doc, event))  # type: ignore
        ]
        self.assertEqual(sorted(completions), ["hello", "hey"])

    def test_prompt_message_multiline(self):
        input_prompt = InputPrompt(
            message="Enter your name",
            style=None,
            default="",
            qmark="[?]",
            vi_mode=False,
            multiline=True,
            completer={"hello": None, "hey": None, "what": None},
        )
        message = input_prompt._get_prompt_message()
        self.assertEqual(
            message,
            [
                ("class:questionmark", "[?]"),
                ("class:question", " Enter your name"),
                ("class:instruction", " ESC + Enter to finish input"),
                ("class:questionmark", "\n%s " % INQUIRERPY_POINTER_SEQUENCE),
            ],
        )
        input_prompt.status["answered"] = True
        input_prompt.status["result"] = "haha\n123"
        message = input_prompt._get_prompt_message()
        self.assertEqual(
            message,
            [
                ("class:answermark", "?"),
                ("class:answered_question", " Enter your name"),
                ("class:answer", " haha...[3 chars]"),
            ],
        )

    def test_prompt_message_instruction(self):
        input_prompt = InputPrompt(
            message="Enter your name:",
            style=None,
            default="",
            qmark="[?]",
            vi_mode=False,
            instruction="(abc)",
        )
        message = input_prompt._get_prompt_message()
        self.assertEqual(
            message,
            [
                ("class:questionmark", "[?]"),
                ("class:question", " Enter your name:"),
                ("class:instruction", " (abc) "),
            ],
        )

        input_prompt = InputPrompt(
            message="Enter your name:",
            style=None,
            default="",
            qmark="[?]",
            vi_mode=False,
            instruction="(abc)",
            multiline=True,
        )
        message = input_prompt._get_prompt_message()
        self.assertEqual(
            message,
            [
                ("class:questionmark", "[?]"),
                ("class:question", " Enter your name:"),
                ("class:instruction", " (abc) "),
                ("class:questionmark", "\n%s " % INQUIRERPY_POINTER_SEQUENCE),
            ],
        )

    @patch("InquirerPy.prompts.input.NestedCompleter.from_nested_dict")
    @patch("InquirerPy.prompts.input.SimpleLexer")
    @patch("InquirerPy.prompts.input.InputPrompt._get_prompt_message")
    @patch("InquirerPy.base.simple.Style.from_dict")
    @patch("InquirerPy.base.simple.KeyBindings")
    @patch("InquirerPy.prompts.input.PromptSession")
    def test_callable_called(
        self,
        MockedSession,
        MockedKeyBindings,
        MockedStyle,
        mocked_message,
        MockedLexer,
        mocked_completer,
    ):
        completer = mocked_completer()
        kb = MockedKeyBindings()
        style = MockedStyle()
        lexer = MockedLexer()
        InputPrompt(
            message="Enter your name",
            style=None,
            default="",
            qmark="[?]",
            vi_mode=False,
            multiline=True,
            completer={"hello": None, "hey": None, "what": None},
        )
        MockedSession.assert_called_once_with(
            message=mocked_message,
            key_bindings=kb,
            style=style,
            completer=completer,
            validator=ANY,
            validate_while_typing=False,
            input=None,
            output=None,
            editing_mode=EditingMode.EMACS,
            lexer=lexer,
            is_password=False,
            multiline=True,
            complete_style=CompleteStyle.COLUMN,
            wrap_lines=True,
            bottom_toolbar=None,
        )
        mocked_completer.assert_has_calls(
            [call({"hello": None, "hey": None, "what": None})]
        )
        MockedLexer.assert_has_calls([call("class:input")])

    @patch("InquirerPy.prompts.input.PromptSession")
    def test_long_instruction(self, MockedSession):
        InputPrompt(message="", long_instruction="asfasdf")

        MockedSession.assert_called_once_with(
            message=ANY,
            key_bindings=ANY,
            style=ANY,
            completer=ANY,
            validator=ANY,
            validate_while_typing=False,
            input=None,
            output=None,
            editing_mode=EditingMode.EMACS,
            lexer=ANY,
            is_password=False,
            multiline=False,
            complete_style=CompleteStyle.COLUMN,
            wrap_lines=True,
            bottom_toolbar=[("class:long_instruction", "asfasdf")],
        )

    def test_handle_completion(self):
        prompt = InputPrompt(message="")
        with patch("prompt_toolkit.utils.Event") as mock:
            event = mock.return_value
            prompt._handle_completion(event)
            mock.assert_not_called()

        prompt = InputPrompt(message="", completer={})
        with patch("prompt_toolkit.utils.Event") as mock:
            event = mock.return_value
            prompt._handle_completion(event)
            mock.assert_has_calls(
                [
                    call().app.current_buffer.complete_state.__bool__(),
                    call().app.current_buffer.complete_next(),
                ]
            )

    def test_prompt_result_async(self):
        self.inp.send_text("hello\n")
        input_prompt = InputPrompt(
            message="yes",
            style=None,
            default="world",
            qmark="!",
            vi_mode=False,
            input=self.inp,
            output=DummyOutput(),
        )
        result = asyncio.run(input_prompt.execute_async())
        self.assertEqual(result, "worldhello")
        self.assertEqual(input_prompt.status["answered"], True)
        self.assertEqual(input_prompt.status["result"], "worldhello")
        self.assertEqual(input_prompt.status["skipped"], False)



import unittest
from unittest.mock import patch

from prompt_toolkit.enums import EditingMode
from prompt_toolkit.key_binding.key_bindings import KeyBindings
from prompt_toolkit.styles.style import Style

from InquirerPy.prompts.list import InquirerPyListControl, ListPrompt
from InquirerPy.separator import Separator
from InquirerPy.utils import InquirerPyStyle


class TestBaseList(unittest.TestCase):
    choices = [
        {"name": "apple", "value": "peach"},
        "pear",
        {"name": "melon", "value": "watermelon"},
    ]

    @patch("InquirerPy.base.complex.shutil.get_terminal_size")
    def test_list_prompt(self, mocked_term):
        mocked_term.return_value = (24, 80)
        message = 15 * "i"
        qmark = "[?]"
        instruction = 2 * "i"
        prompt = ListPrompt(
            message=message,
            choices=self.choices,
            default="watermelon",
            style=InquirerPyStyle({"pointer": "#61afef"}),
            vi_mode=True,
            qmark=qmark,
            pointer=">",
            instruction=instruction,
            show_cursor=True,
            wrap_lines=True,
            border=True,
        )
        self.assertEqual(prompt._editing_mode, EditingMode.VI)
        self.assertIsInstance(prompt.content_control, InquirerPyListControl)
        self.assertIsInstance(prompt._kb, KeyBindings)
        self.assertIsInstance(prompt._style, Style)
        self.assertEqual(prompt._message, message)
        self.assertEqual(prompt._qmark, qmark)
        self.assertEqual(prompt.instruction, instruction)
        self.assertTrue(prompt._border, True)

    def test_prompt_message_with_cursor(self):
        prompt = ListPrompt(message="Select one:", choices=[1, 2, 3])
        self.assertEqual(
            prompt._get_prompt_message_with_cursor(),
            [
                ("class:questionmark", "?"),
                ("class:question", " Select one:"),
                ("class:instruction", " "),
                ("[SetCursorPosition]", ""),
                ("", " "),
            ],
        )

    def test_prompt_handle_toggle_choice_no_multiselect(self):
        prompt = ListPrompt(message="Select one:", choices=[1, 2, 3], multiselect=False)
        self.assertEqual(
            prompt.content_control.selection,
            {"enabled": False, "name": "1", "value": 1},
        )
        prompt._handle_toggle_choice(None)
        self.assertEqual(
            prompt.content_control.selection,
            {"enabled": False, "name": "1", "value": 1},
        )

    def test_prompt_handle_toggle_choice(self):
        prompt = ListPrompt(message="Select one:", choices=[1, 2, 3], multiselect=True)
        self.assertEqual(
            prompt.content_control.selection,
            {"enabled": False, "name": "1", "value": 1},
        )
        prompt._handle_toggle_choice(None)
        self.assertEqual(
            prompt.content_control.selection,
            {"enabled": True, "name": "1", "value": 1},
        )
        prompt._handle_toggle_choice(None)
        self.assertEqual(
            prompt.content_control.selection,
            {"enabled": False, "name": "1", "value": 1},
        )

    def test_prompt_handle_toggle_all_no_multiselect(self) -> None:
        prompt = ListPrompt(message="Select one:", choices=[1, 2, 3])
        self.assertEqual(
            prompt.content_control.choices,
            [
                {"enabled": False, "name": "1", "value": 1},
                {"enabled": False, "name": "2", "value": 2},
                {"enabled": False, "name": "3", "value": 3},
            ],
        )
        prompt._handle_toggle_all(None)
        self.assertEqual(
            prompt.content_control.choices,
            [
                {"enabled": False, "name": "1", "value": 1},
                {"enabled": False, "name": "2", "value": 2},
                {"enabled": False, "name": "3", "value": 3},
            ],
        )

    def test_prompt_handle_toggle_all(self):
        prompt = ListPrompt(message="Select one:", choices=[1, 2, 3], multiselect=True)
        self.assertEqual(
            prompt.content_control.choices,
            [
                {"enabled": False, "name": "1", "value": 1},
                {"enabled": False, "name": "2", "value": 2},
                {"enabled": False, "name": "3", "value": 3},
            ],
        )
        prompt._handle_toggle_all(None)
        self.assertEqual(
            prompt.content_control.choices,
            [
                {"enabled": True, "name": "1", "value": 1},
                {"enabled": True, "name": "2", "value": 2},
                {"enabled": True, "name": "3", "value": 3},
            ],
        )
        prompt._handle_toggle_all(None)
        self.assertEqual(
            prompt.content_control.choices,
            [
                {"enabled": False, "name": "1", "value": 1},
                {"enabled": False, "name": "2", "value": 2},
                {"enabled": False, "name": "3", "value": 3},
            ],
        )

    def test_handle_up(self):
        prompt = ListPrompt(message="Select one:", choices=[1, 2, Separator(), 3])
        self.assertEqual(prompt.content_control.selected_choice_index, 0)
        prompt._handle_up(None)
        self.assertEqual(prompt.content_control.selected_choice_index, 3)
        prompt._handle_up(None)
        self.assertEqual(prompt.content_control.selected_choice_index, 1)
        prompt._handle_up(None)
        self.assertEqual(prompt.content_control.selected_choice_index, 0)
        prompt._handle_up(None)
        self.assertEqual(prompt.content_control.selected_choice_index, 3)

        prompt = ListPrompt(message="Select one:", choices=[1, 2, 3, Separator()])
        self.assertEqual(prompt.content_control.selected_choice_index, 0)
        prompt._handle_up(None)
        self.assertEqual(prompt.content_control.selected_choice_index, 2)

    def test_handle_up_no_cycle(self):
        prompt = ListPrompt(
            message="Select one:",
            choices=[Separator(), 1, 2, 3, Separator()],
            cycle=False,
        )
        self.assertEqual(prompt.content_control.selected_choice_index, 1)
        prompt._handle_up(None)
        self.assertEqual(prompt.content_control.selected_choice_index, 1)

        prompt = ListPrompt(
            message="Select one:",
            choices=[1, 2, 3, Separator()],
            cycle=False,
        )
        self.assertEqual(prompt.content_control.selected_choice_index, 0)
        prompt._handle_up(None)
        self.assertEqual(prompt.content_control.selected_choice_index, 0)

    def test_handle_down(self) -> None:
        prompt = ListPrompt(
            message="Select one:",
            choices=[1, 2, 3, Separator()],
        )
        self.assertEqual(prompt.content_control.selected_choice_index, 0)
        prompt._handle_down(None)
        self.assertEqual(prompt.content_control.selected_choice_index, 1)
        prompt._handle_down(None)
        prompt._handle_down(None)
        self.assertEqual(prompt.content_control.selected_choice_index, 0)

    def test_handle_down_no_cycle(self):
        prompt = ListPrompt(
            message="Select one:",
            choices=[Separator(), 1, 2, 3, Separator()],
            cycle=False,
        )
        self.assertEqual(prompt.content_control.selected_choice_index, 1)
        prompt._handle_down(None)
        self.assertEqual(prompt.content_control.selected_choice_index, 2)
        prompt._handle_down(None)
        prompt._handle_down(None)
        self.assertEqual(prompt.content_control.selected_choice_index, 3)

        prompt = ListPrompt(
            message="Select one:",
            choices=[1, 2, 3, Separator()],
            cycle=False,
        )
        self.assertEqual(prompt.content_control.selected_choice_index, 0)
        prompt._handle_down(None)
        prompt._handle_down(None)
        prompt._handle_down(None)
        prompt._handle_down(None)
        self.assertEqual(prompt.content_control.selected_choice_index, 2)

    def test_handle_enter(self) -> None:
        prompt = ListPrompt(
            message="Select a fruit",
            choices=self.choices,
            default="watermelon",
            style=InquirerPyStyle({"pointer": "#61afef"}),
            vi_mode=True,
            qmark="[?]",
            pointer=">",
            instruction="(j/k)",
        )
        self.assertEqual(
            prompt.status, {"result": None, "answered": False, "skipped": False}
        )
        with patch("prompt_toolkit.utils.Event") as mock:
            event = mock.return_value
            prompt._handle_enter(event)
        self.assertEqual(
            prompt.status, {"result": "melon", "answered": True, "skipped": False}
        )

    def test_handle_enter_multi(self) -> None:
        prompt = ListPrompt(
            message="Select a fruit",
            choices=self.choices,
            default="watermelon",
            style=InquirerPyStyle({"pointer": "#61afef"}),
            vi_mode=True,
            qmark="[?]",
            pointer=">",
            instruction="(j/k)",
            multiselect=True,
        )
        self.assertEqual(
            prompt.status, {"result": None, "answered": False, "skipped": False}
        )
        with patch("prompt_toolkit.utils.Event") as mock:
            event = mock.return_value
            prompt._handle_enter(event)
        self.assertEqual(
            prompt.status, {"result": ["melon"], "answered": True, "skipped": False}
        )

    def test_handle_enter_validator(self):
        prompt = ListPrompt(
            message="Select a fruit",
            choices=self.choices,
            default="watermelon",
            style=InquirerPyStyle({"pointer": "#61afef"}),
            vi_mode=True,
            qmark="[?]",
            pointer=">",
            instruction="(j/k)",
            validate=lambda result: result != "watermelon",
        )
        self.assertFalse(prompt._invalid)
        self.assertEqual(
            prompt.status, {"result": None, "answered": False, "skipped": False}
        )
        with patch("prompt_toolkit.utils.Event") as mock:
            event = mock.return_value
            prompt._handle_enter(event)
        self.assertEqual(
            prompt.status, {"result": None, "answered": False, "skipped": False}
        )
        self.assertTrue(prompt._invalid)

        prompt.content_control.selected_choice_index = 0
        with patch("prompt_toolkit.utils.Event") as mock:
            event = mock.return_value
            prompt._handle_enter(event)
        self.assertEqual(
            prompt.status, {"result": "apple", "answered": True, "skipped": False}
        )

    @patch("InquirerPy.base.complex.shutil.get_terminal_size")
    def test_wrap_lines_offset(self, mocked_term):
        mocked_term.return_value = (24, 80)
        message = 15 * "i"
        qmark = "[?]"
        instruction = 3 * "i"
        prompt = ListPrompt(
            message=message,
            choices=self.choices,
            default="watermelon",
            style=InquirerPyStyle({"pointer": "#61afef"}),
            vi_mode=True,
            qmark=qmark,
            pointer=">",
            instruction=instruction,
            show_cursor=True,
            wrap_lines=True,
        )
        self.assertEqual(
            prompt.extra_line_count,
            (len(qmark) + 1 + len(message) + 1 + len(instruction) + 1) // 24,
        )

        prompt = ListPrompt(
            message=message,
            choices=self.choices,
            default="watermelon",
            style=InquirerPyStyle({"pointer": "#61afef"}),
            vi_mode=True,
            qmark=qmark,
            pointer=">",
            instruction=instruction,
            show_cursor=False,
            wrap_lines=True,
        )
        self.assertEqual(
            prompt.extra_line_count,
            (len(qmark) + 1 + len(message) + 1 + len(instruction) + 1 - 1) // 24,
        )
        prompt._wrap_lines = False
        self.assertEqual(prompt.extra_line_count, 0)



"""Module contains the class to create a rawlist prompt."""
from typing import Any, Callable, List, Optional, Tuple, Union

from InquirerPy.base import InquirerPyUIListControl
from InquirerPy.enum import INQUIRERPY_POINTER_SEQUENCE
from InquirerPy.exceptions import InvalidArgument
from InquirerPy.prompts.list import ListPrompt
from InquirerPy.separator import Separator
from InquirerPy.utils import (
    InquirerPyDefault,
    InquirerPyKeybindings,
    InquirerPyListChoices,
    InquirerPyMessage,
    InquirerPySessionResult,
    InquirerPyStyle,
    InquirerPyValidate,
)

__all__ = ["RawlistPrompt"]


class InquirerPyRawlistControl(InquirerPyUIListControl):
    """An :class:`~prompt_toolkit.layout.UIControl` class that displays a list of choices.

    Reference the parameter definition in :class:`.RawlistPrompt`.
    """

    def __init__(
        self,
        choices: InquirerPyListChoices,
        default: Any,
        pointer: str,
        separator: str,
        marker: str,
        session_result: Optional[InquirerPySessionResult],
        multiselect: bool,
        marker_pl: str,
    ) -> None:
        self._pointer = pointer
        self._separator = separator
        self._marker = marker
        self._marker_pl = marker_pl
        super().__init__(
            choices=choices,
            default=default,
            session_result=session_result,
            multiselect=multiselect,
        )

    def _format_choices(self) -> None:
        separator_count = 0
        for index, choice in enumerate(self.choices):
            if isinstance(choice["value"], Separator):
                separator_count += 1
                continue
            choice["display_index"] = index + 1 - separator_count
            choice["actual_index"] = index

        if self.choices:
            first_valid_choice_index = 0
            while isinstance(
                self.choices[first_valid_choice_index]["value"], Separator
            ):
                first_valid_choice_index += 1
            if self.selected_choice_index == first_valid_choice_index:
                for choice in self.choices:
                    if isinstance(choice["value"], Separator):
                        continue
                    if choice["display_index"] == self._default:
                        self.selected_choice_index = choice["actual_index"]
                        break

    def _get_hover_text(self, choice) -> List[Tuple[str, str]]:
        display_choices = []
        display_choices.append(("class:pointer", self._pointer))
        display_choices.append(
            (
                "class:marker",
                self._marker if choice["enabled"] else self._marker_pl,
            )
        )
        if not isinstance(choice["value"], Separator):
            display_choices.append(
                (
                    "class:pointer",
                    "%s%s" % (str(choice["display_index"]), self._separator),
                )
            )
        display_choices.append(("[SetCursorPosition]", ""))
        display_choices.append(("class:pointer", choice["name"]))
        return display_choices

    def _get_normal_text(self, choice) -> List[Tuple[str, str]]:
        display_choices = []
        display_choices.append(("", len(self._pointer) * " "))
        display_choices.append(
            (
                "class:marker",
                self._marker if choice["enabled"] else self._marker_pl,
            )
        )
        if not isinstance(choice["value"], Separator):
            display_choices.append(
                ("", "%s%s" % (str(choice["display_index"]), self._separator))
            )
            display_choices.append(("", choice["name"]))
        else:
            display_choices.append(("class:separator", choice["name"]))
        return display_choices


class RawlistPrompt(ListPrompt):
    """Create a prompt that displays a list of choices with index number as shortcuts.

    A wrapper class around :class:`~prompt_toolkit.application.Application`.

    Each choice will have an index number infront of them with keybinding created for that index number.
    Enables user to use number to jump to different choices using number.

    Args:
        message: The question to ask the user.
            Refer to :ref:`pages/dynamic:message` documentation for more details.
        choices: List of choices to display and select.
            Refer to :ref:`pages/prompts/rawlist:Choices` documentation for more details.
        style: An :class:`InquirerPyStyle` instance.
            Refer to :ref:`Style <pages/style:Alternate Syntax>` documentation for more details.
        vi_mode: Use vim keybinding for the prompt.
            Refer to :ref:`pages/kb:Keybindings` documentation for more details.
        default: Set the default value of the prompt.
            This will be used to determine which choice is highlighted (current selection),
            The default value should be the value of one of the choices.
            For :class:`.RawlistPrompt` specifically, default value can also be value between 0-9.
            Refer to :ref:`pages/dynamic:default` documentation for more details.
        separator: Separator symbol. Custom symbol that will be used as a separator between the choice index number and the choices.
        qmark: Question mark symbol. Custom symbol that will be displayed infront of the question before its answered.
        amark: Answer mark symbol. Custom symbol that will be displayed infront of the question after its answered.
        pointer: Pointer symbol. Customer symbol that will be used to indicate the current choice selection.
        instruction: Short instruction to display next to the question.
        long_instruction: Long instructions to display at the bottom of the prompt.
        validate: Add validation to user input.
            The main use case for this prompt would be when `multiselect` is True, you can enforce a min/max selection.
            Refer to :ref:`pages/validator:Validator` documentation for more details.
        invalid_message: Error message to display when user input is invalid.
            Refer to :ref:`pages/validator:Validator` documentation for more details.
        transformer: A function which performs additional transformation on the value that gets printed to the terminal.
            Different than `filter` parameter, this is only visual effect and won’t affect the actual value returned by :meth:`~InquirerPy.base.simple.BaseSimplePrompt.execute`.
            Refer to :ref:`pages/dynamic:transformer` documentation for more details.
        filter: A function which performs additional transformation on the result.
            This affects the actual value returned by :meth:`~InquirerPy.base.simple.BaseSimplePrompt.execute`.
            Refer to :ref:`pages/dynamic:filter` documentation for more details.
        height: Preferred height of the prompt.
            Refer to :ref:`pages/height:Height` documentation for more details.
        max_height: Max height of the prompt.
            Refer to :ref:`pages/height:Height` documentation for more details.
        multiselect: Enable multi-selection on choices.
            You can use `validate` parameter to control min/max selections.
            Setting to True will also change the result from a single value to a list of values.
        marker: Marker Symbol. Custom symbol to indicate if a choice is selected.
            This will take effects when `multiselect` is True.
        marker_pl: Marker place holder when the choice is not selected.
            This is empty space by default.
        border: Create border around the choice window.
        keybindings: Customise the builtin keybindings.
            Refer to :ref:`pages/kb:Keybindings` for more details.
        show_cursor: Display cursor at the end of the prompt.
            Set to False to hide the cursor.
        cycle: Return to top item if hit bottom during navigation or vice versa.
        wrap_lines: Soft wrap question lines when question exceeds the terminal width.
        raise_keyboard_interrupt: Raise the :class:`KeyboardInterrupt` exception when `ctrl-c` is pressed. If false, the result
            will be `None` and the question is skiped.
        mandatory: Indicate if the prompt is mandatory. If True, then the question cannot be skipped.
        mandatory_message: Error message to show when user attempts to skip mandatory prompt.
        session_result: Used internally for :ref:`index:Classic Syntax (PyInquirer)`.

    Examples:
        >>> from InquirerPy import inquirer
        >>> result = inquirer.rawlist(message="Select one:", choices=[1, 2, 3]).execute()
        >>> print(result)
        1
    """

    def __init__(
        self,
        message: InquirerPyMessage,
        choices: InquirerPyListChoices,
        default: InquirerPyDefault = None,
        separator: str = ") ",
        style: Optional[InquirerPyStyle] = None,
        vi_mode: bool = False,
        qmark: str = "?",
        amark: str = "?",
        pointer: str = " ",
        instruction: str = "",
        long_instruction: str = "",
        transformer: Optional[Callable[[Any], Any]] = None,
        filter: Optional[Callable[[Any], Any]] = None,
        height: Optional[Union[int, str]] = None,
        max_height: Optional[Union[int, str]] = None,
        multiselect: bool = False,
        marker: str = INQUIRERPY_POINTER_SEQUENCE,
        marker_pl: str = " ",
        border: bool = False,
        validate: Optional[InquirerPyValidate] = None,
        invalid_message: str = "Invalid input",
        keybindings: Optional[InquirerPyKeybindings] = None,
        show_cursor: bool = True,
        cycle: bool = True,
        wrap_lines: bool = True,
        raise_keyboard_interrupt: bool = True,
        mandatory: bool = True,
        mandatory_message: str = "Mandatory prompt",
        session_result: Optional[InquirerPySessionResult] = None,
    ) -> None:
        self.content_control = InquirerPyRawlistControl(
            choices=choices,
            default=default,
            pointer=pointer,
            separator=separator,
            marker=marker,
            session_result=session_result,
            multiselect=multiselect,
            marker_pl=marker_pl,
        )
        super().__init__(
            message=message,
            choices=choices,
            style=style,
            border=border,
            vi_mode=vi_mode,
            qmark=qmark,
            amark=amark,
            instruction=instruction,
            long_instruction=long_instruction,
            transformer=transformer,
            filter=filter,
            height=height,
            max_height=max_height,
            multiselect=multiselect,
            validate=validate,
            invalid_message=invalid_message,
            keybindings=keybindings,
            show_cursor=show_cursor,
            cycle=cycle,
            wrap_lines=wrap_lines,
            raise_keyboard_interrupt=raise_keyboard_interrupt,
            mandatory=mandatory,
            mandatory_message=mandatory_message,
            session_result=session_result,
        )

    def _on_rendered(self, _) -> None:
        """Override this method to apply custom keybindings.

        Needs to creat these kb in the callback due to `after_render`
        retrieve the choices asynchronously.

        Check if fetched choices exceed the limit of 9, raise
        InvalidArgument exception.
        """

        def keybinding_factory(choice):
            @self.register_kb(str(choice["display_index"]))
            def keybinding(_) -> None:
                self.content_control.selected_choice_index = int(choice["actual_index"])

            return keybinding

        if self.content_control.choice_count >= 10:
            raise InvalidArgument("rawlist argument choices cannot exceed length of 9")

        for choice in self.content_control.choices:
            if not isinstance(choice["value"], Separator):
                keybinding_factory(choice)

    def _get_prompt_message(self) -> List[Tuple[str, str]]:
        """Return the formatted text to display in the prompt.

        Overriding this method to allow multiple formatted class to be displayed.
        """
        display_message = super()._get_prompt_message()
        if not self.status["answered"] and self.content_control.choices:
            display_message.append(
                ("class:input", str(self.content_control.selection["display_index"]))
            )
        return display_message



import asyncio
import os
import sys
import unittest
from unittest.mock import ANY, call, patch

from prompt_toolkit.shortcuts.prompt import PromptSession

from InquirerPy.base.simple import BaseSimplePrompt
from InquirerPy.enum import INQUIRERPY_KEYBOARD_INTERRUPT
from InquirerPy.exceptions import InvalidArgument, RequiredKeyNotFound
from InquirerPy.prompts import FuzzyPrompt
from InquirerPy.prompts.confirm import ConfirmPrompt
from InquirerPy.prompts.expand import ExpandPrompt
from InquirerPy.prompts.filepath import FilePathPrompt
from InquirerPy.prompts.input import InputPrompt
from InquirerPy.prompts.list import ListPrompt
from InquirerPy.prompts.secret import SecretPrompt
from InquirerPy.resolver import prompt, prompt_async
from InquirerPy.utils import InquirerPyStyle

from .style import get_sample_style


class TestResolver(unittest.TestCase):
    @patch("InquirerPy.resolver.ConfirmPrompt.execute")
    def test_exceptions(self, mocked_confirm):
        questions = "hello"
        self.assertRaises(InvalidArgument, prompt, questions)

        questions = [{"name": "hello"}]
        self.assertRaises(RequiredKeyNotFound, prompt, questions)

        questions = [{"type": "confirm"}]
        self.assertRaises(RequiredKeyNotFound, prompt, questions)

        questions = [{"type": "confirm", "message": "hello"}]
        mocked_confirm.return_value = True
        result = prompt(questions)
        mocked_confirm.assert_called_once()
        self.assertEqual(result, {0: True})

    @patch.object(FilePathPrompt, "__init__")
    @patch.object(FilePathPrompt, "execute")
    @patch.object(ConfirmPrompt, "__init__")
    @patch.object(ConfirmPrompt, "execute")
    def test_resolver_normal(
        self,
        mocked_confirm_execute,
        mocked_confirm_init,
        mocked_filepath_execute,
        mocked_filepath_init,
    ):
        mocked_confirm_init.return_value = None
        mocked_confirm_execute.return_value = False
        questions = [
            {"type": "confirm", "message": "hello"},
        ]
        result = prompt(questions)
        mocked_confirm_init.assert_called_once_with(
            message="hello",
            style=InquirerPyStyle(get_sample_style()),
            vi_mode=False,
            raise_keyboard_interrupt=True,
            session_result=ANY,
            keybindings={},
        )
        mocked_confirm_execute.assert_called_once()
        self.assertEqual(result, {0: False})

        mocked_filepath_init.return_value = None
        mocked_filepath_execute.return_value = "hello.py"
        mocked_confirm_init.reset_mock()
        mocked_confirm_execute.reset_mock()
        questions = [
            {"type": "confirm", "message": "hello"},
            {"type": "confirm", "message": "world", "name": "foo"},
            {"type": "filepath", "message": "whaat", "name": "boo", "default": "./"},
        ]
        result = prompt(questions)
        mocked_confirm_init.assert_has_calls(
            [
                call(
                    message="hello",
                    style=InquirerPyStyle(get_sample_style()),
                    vi_mode=False,
                    raise_keyboard_interrupt=True,
                    session_result=ANY,
                    keybindings={},
                ),
                call(
                    message="world",
                    style=InquirerPyStyle(get_sample_style()),
                    vi_mode=False,
                    raise_keyboard_interrupt=True,
                    session_result=ANY,
                    keybindings={},
                ),
            ]
        )
        mocked_filepath_init.assert_has_calls(
            [
                call(
                    message="whaat",
                    style=InquirerPyStyle(get_sample_style()),
                    default="./",
                    vi_mode=False,
                    raise_keyboard_interrupt=True,
                    session_result=ANY,
                    keybindings={},
                )
            ]
        )
        mocked_confirm_execute.assert_has_calls([call(), call()])
        self.assertEqual(result, {0: False, "foo": False, "boo": "hello.py"})

    @patch.object(SecretPrompt, "__init__")
    @patch.object(SecretPrompt, "execute")
    @patch.object(ConfirmPrompt, "__init__")
    @patch.object(ConfirmPrompt, "execute")
    def test_resolver_style_keys(
        self,
        mocked_confirm_execute,
        mocked_confirm_init,
        mocked_secret_execute,
        mocked_secret_init,
    ):
        mocked_confirm_execute.return_value = False
        mocked_confirm_init.return_value = None
        os.environ["INQUIRERPY_VI_MODE"] = "true"

        questions = [{"type": "confirm", "message": "Confirm?", "name": "question1"}]
        result = prompt(questions)
        mocked_confirm_execute.assert_called_once()
        mocked_confirm_init.assert_called_once_with(
            message="Confirm?",
            style=InquirerPyStyle(get_sample_style()),
            vi_mode=False,
            raise_keyboard_interrupt=True,
            session_result={"question1": False},
            keybindings={},
        )
        self.assertEqual(result, {"question1": False})
        del os.environ["INQUIRERPY_VI_MODE"]

        mocked_secret_init.return_value = None
        mocked_secret_execute.return_value = "111111"
        mocked_confirm_execute.reset_mock()
        mocked_confirm_init.reset_mock()
        mocked_confirm_execute.return_value = True
        questions = [
            {"type": "confirm", "message": "Confirm?", "name": "10"},
            {"type": "confirm", "message": "What?"},
            {"type": "password", "message": "haha"},
        ]
        result = prompt(
            questions, style={"qmark": "#ffffff"}, vi_mode=True, style_override=False
        )
        mocked_confirm_execute.assert_has_calls([call(), call()])
        mocked_confirm_init.assert_has_calls(
            [
                call(
                    message="Confirm?",
                    style=InquirerPyStyle(get_sample_style({"qmark": "#ffffff"})),
                    vi_mode=True,
                    raise_keyboard_interrupt=True,
                    session_result=ANY,
                    keybindings={},
                ),
                call(
                    message="What?",
                    style=InquirerPyStyle(get_sample_style({"qmark": "#ffffff"})),
                    vi_mode=True,
                    raise_keyboard_interrupt=True,
                    session_result=ANY,
                    keybindings={},
                ),
            ]
        )
        mocked_secret_init.reset_mock()
        questions = [
            {"type": "confirm", "message": "Confirm?", "name": "10"},
            {"type": "confirm", "message": "What?"},
            {"type": "password", "message": "haha"},
        ]
        result = prompt(
            questions,
            style={"qmark": "#ffffff"},
            vi_mode=True,
            style_override=True,
        )
        style = get_sample_style()
        for key in style.keys():
            style[key] = ""
        style["qmark"] = "#ffffff"
        style["fuzzy_border"] = style.pop("frame.border")
        style["bottom-toolbar"] = "noreverse"
        mocked_secret_init.assert_has_calls(
            [
                call(
                    message="haha",
                    style=InquirerPyStyle(style),
                    vi_mode=True,
                    raise_keyboard_interrupt=True,
                    session_result={"10": True, 1: True, 2: "111111"},
                    keybindings={},
                )
            ]
        )
        self.assertEqual(result, {"10": True, 1: True, 2: "111111"})

    @patch.object(ConfirmPrompt, "__init__")
    @patch.object(ConfirmPrompt, "execute")
    def test_resolver_condition(self, mocked_execute, mocked_init):
        mocked_init.return_value = None
        mocked_execute.return_value = True
        questions = [
            {
                "type": "confirm",
                "name": "first",
                "message": "Confirm first?",
                "default": True,
            },
            {
                "type": "confirm",
                "name": "second",
                "message": "Confirm second?",
                "when": lambda result: result["first"] == True,
            },
            {
                "type": "confirm",
                "name": "third",
                "message": "Confirm?",
                "when": lambda result: result["second"] == False,
            },
        ]
        result = prompt(questions)
        mocked_execute.assert_has_calls([call(), call()])
        mocked_init.assert_has_calls(
            [
                call(
                    message="Confirm first?",
                    style=InquirerPyStyle(get_sample_style()),
                    vi_mode=False,
                    default=True,
                    raise_keyboard_interrupt=True,
                    session_result=ANY,
                    keybindings={},
                ),
                call(
                    message="Confirm second?",
                    style=InquirerPyStyle(get_sample_style()),
                    vi_mode=False,
                    raise_keyboard_interrupt=True,
                    session_result=ANY,
                    keybindings={},
                ),
            ]
        )
        self.assertEqual(result, {"first": True, "second": True, "third": None})

        mocked_init.reset_mock()
        mocked_execute.reset_mock()
        questions = [
            {
                "type": "confirm",
                "name": "first",
                "message": "Confirm first?",
                "default": True,
            },
            {
                "type": "confirm",
                "name": "second",
                "message": "Confirm second?",
                "when": lambda result: result["first"] == False,
            },
            {
                "type": "confirm",
                "name": "third",
                "message": "Confirm?",
                "when": lambda result: result["second"] == None,
            },
        ]
        result = prompt(questions)
        mocked_execute.assert_has_calls([call(), call()])
        mocked_init.assert_has_calls(
            [
                call(
                    message="Confirm first?",
                    style=InquirerPyStyle(get_sample_style()),
                    vi_mode=False,
                    default=True,
                    raise_keyboard_interrupt=True,
                    session_result=ANY,
                    keybindings={},
                ),
                call(
                    message="Confirm?",
                    style=InquirerPyStyle(get_sample_style()),
                    vi_mode=False,
                    raise_keyboard_interrupt=True,
                    session_result=ANY,
                    keybindings={},
                ),
            ]
        )
        self.assertEqual(result, {"first": True, "second": None, "third": True})

    @patch.object(ExpandPrompt, "execute")
    def test_resolver_filter(self, mocked_execute):
        mocked_execute.return_value = "boo"
        prompt = ExpandPrompt(
            message="Select one",
            choices=[
                {"name": "foo", "value": "boo", "key": "f"},
                {"name": "hello", "value": "world", "key": "w"},
            ],
            filter=lambda x: 2 * x,
        )
        result = prompt._filter(prompt.execute())
        self.assertEqual(result, "booboo")

    def test_resolver_transformer(self):
        prompt = ListPrompt(
            message="Select one", choices=["1", "2", "3"], transformer=lambda x: x * 2
        )
        self.assertEqual(
            prompt._get_prompt_message(),
            [
                ("class:questionmark", "?"),
                ("class:question", " Select one"),
                ("class:instruction", " "),
            ],
        )
        prompt.status["result"] = "1"
        prompt.status["answered"] = True
        self.assertEqual(
            prompt._get_prompt_message(),
            [
                ("class:answermark", "?"),
                ("class:answered_question", " Select one"),
                ("class:answer", " 11"),
            ],
        )

        prompt = InputPrompt(
            message="Select one",
            transformer=lambda x: x * 2,
        )
        prompt.status["result"] = "2"
        prompt.status["answered"] = True
        self.assertEqual(
            prompt._get_prompt_message(),
            [
                ("class:answermark", "?"),
                ("class:answered_question", " Select one"),
                ("class:answer", " 22"),
            ],
        )

        prompt = SecretPrompt(
            message="Select one",
            transformer=lambda x: x * 2,
        )
        prompt.status["result"] = "2"
        prompt.status["answered"] = True
        self.assertEqual(
            prompt._get_prompt_message(),
            [
                ("class:answermark", "?"),
                ("class:answered_question", " Select one"),
                ("class:answer", " 22"),
            ],
        )

    @patch.object(PromptSession, "prompt")
    def test_optional_keyboard_interrupt(self, mocked_execute):
        mocked_execute.return_value = INQUIRERPY_KEYBOARD_INTERRUPT
        questions = [{"type": "input", "message": "hello"}]
        try:
            prompt(questions, raise_keyboard_interrupt=False)
        except KeyboardInterrupt:
            pass
        else:
            self.fail("should raise kbi")

        mocked_execute.return_value = INQUIRERPY_KEYBOARD_INTERRUPT
        self.assertRaises(
            KeyboardInterrupt, prompt, questions, raise_keyboard_interrupt=True
        )

        input_prompt = InputPrompt(message="hello")
        input_prompt.status["answered"] = True
        input_prompt.status["result"] = INQUIRERPY_KEYBOARD_INTERRUPT
        input_prompt.status["skipped"] = True
        self.assertEqual(
            input_prompt._get_prompt_message(),
            [("class:skipped", "?"), ("class:skipped", " hello ")],
        )

    @patch.object(ListPrompt, "execute")
    @patch.object(InputPrompt, "execute")
    @patch.object(BaseSimplePrompt, "register_kb")
    def test_custom_kb(self, mocked_kb, mocked_execute1, mocked_execute2):
        questions = [{"type": "input", "message": "hello"}]
        prompt(questions, keybindings={"answer": [{"key": "up"}]})
        mocked_kb.assert_has_calls([call("up", filter=ANY)])
        try:
            mocked_kb.assert_has_calls([call("enter", filter=ANY)])
            self.fail("enter should not have been bound")
        except:
            pass

        mocked_kb.reset_mock()
        questions = [
            {
                "type": "list",
                "message": "aasdf",
                "choices": [1, 2, 3],
                "keybindings": {"up": [{"key": "c-w"}]},
            }
        ]
        prompt(questions, keybindings={"up": [{"key": "c-p"}]}, vi_mode=True)
        try:
            mocked_kb.assert_has_calls([call("c-p", filter=ANY)])
            self.fail("should not have called")
        except:
            pass

    @patch.object(FuzzyPrompt, "__init__")
    @patch.object(FuzzyPrompt, "execute")
    def test_resolver_when_not_poped(self, mocked_execute, mocked_init):
        mocked_init.return_value = None
        mocked_execute.return_value = 1
        questions = [
            {"message": "", "choices": [1, 2, 3], "type": "fuzzy"},
            {
                "message": "",
                "type": "fuzzy",
                "choices": [4, 5, 6],
                "when": lambda result: result == 1,
            },
        ]
        prompt(questions)

    @patch.object(InputPrompt, "execute")
    def test_single_dict_question(self, mocked_execute):
        mocked_execute.return_value = None
        prompt({"type": "input", "message": "Name:"})

    @patch("InquirerPy.resolver.ConfirmPrompt.execute")
    def test_prompt_async_exception(self, mocked_execute):
        questions = "hello"
        try:
            asyncio.run(prompt_async(questions))
        except InvalidArgument:
            pass
        else:
            self.fail("InvalidArgument should be raised")

    @patch("InquirerPy.resolver.ConfirmPrompt.__init__")
    @patch("InquirerPy.resolver.InputPrompt.__init__")
    @patch("InquirerPy.resolver.ConfirmPrompt.execute_async")
    @patch("InquirerPy.resolver.InputPrompt.execute_async")
    def test_prompt_async(
        self, mocked_input, mocked_confirm, mocked_input_init, mocked_confirm_init
    ):
        if sys.version_info.minor == 7:
            return
        mocked_confirm_init.return_value = None
        mocked_input_init.return_value = None
        questions = [
            {"type": "confirm", "message": ""},
            {"type": "input", "message": "", "when": lambda result: result[0]},
        ]
        mocked_confirm.return_value = True
        mocked_input.return_value = "1"
        result = asyncio.run(prompt_async(questions))
        self.assertEqual(result, {0: True, 1: "1"})

        mocked_confirm.reset_mock()
        mocked_confirm.return_value = False
        wtf = asyncio.run(prompt_async(questions))
        self.assertEqual(wtf, {0: False, 1: None})



"""Contains the interface class :class:`.BaseComplexPrompt` for more complex prompts and the mocked document class :class:`.FakeDocument`."""
import shutil
from dataclasses import dataclass
from typing import Any, Callable, List, Optional, Tuple, Union

from prompt_toolkit.application import Application
from prompt_toolkit.enums import EditingMode
from prompt_toolkit.filters.base import Condition, FilterOrBool
from prompt_toolkit.key_binding.key_bindings import KeyHandlerCallable
from prompt_toolkit.keys import Keys

from InquirerPy.base.simple import BaseSimplePrompt
from InquirerPy.enum import INQUIRERPY_KEYBOARD_INTERRUPT
from InquirerPy.utils import (
    InquirerPySessionResult,
    InquirerPyStyle,
    InquirerPyValidate,
)


@dataclass
class FakeDocument:
    """A fake `prompt_toolkit` document class.

    Work around to allow non-buffer type :class:`~prompt_toolkit.layout.UIControl` to use
    :class:`~prompt_toolkit.validation.Validator`.

    Args:
        text: Content to be validated.
        cursor_position: Fake cursor position.
    """

    text: str
    cursor_position: int = 0


class BaseComplexPrompt(BaseSimplePrompt):
    """A base class to create a more complex prompt that will involve :class:`~prompt_toolkit.application.Application`.

    Note:
        This class does not create :class:`~prompt_toolkit.layout.Layout` nor :class:`~prompt_toolkit.application.Application`,
        it only contains the necessary attributes and helper functions to be consumed.

    Note:
        Use :class:`~InquirerPy.base.BaseListPrompt` to create a complex list prompt which involves multiple choices. It has
        more methods and helper function implemented.

    See Also:
        :class:`~InquirerPy.base.BaseListPrompt`
        :class:`~InquirerPy.prompts.fuzzy.FuzzyPrompt`
    """

    def __init__(
        self,
        message: Union[str, Callable[[InquirerPySessionResult], str]],
        style: Optional[InquirerPyStyle] = None,
        border: bool = False,
        vi_mode: bool = False,
        qmark: str = "?",
        amark: str = "?",
        instruction: str = "",
        long_instruction: str = "",
        transformer: Optional[Callable[[Any], Any]] = None,
        filter: Optional[Callable[[Any], Any]] = None,
        validate: Optional[InquirerPyValidate] = None,
        invalid_message: str = "Invalid input",
        wrap_lines: bool = True,
        raise_keyboard_interrupt: bool = True,
        mandatory: bool = True,
        mandatory_message: str = "Mandatory prompt",
        session_result: Optional[InquirerPySessionResult] = None,
    ) -> None:
        super().__init__(
            message=message,
            style=style,
            vi_mode=vi_mode,
            qmark=qmark,
            amark=amark,
            instruction=instruction,
            transformer=transformer,
            filter=filter,
            invalid_message=invalid_message,
            validate=validate,
            wrap_lines=wrap_lines,
            raise_keyboard_interrupt=raise_keyboard_interrupt,
            mandatory=mandatory,
            mandatory_message=mandatory_message,
            session_result=session_result,
        )
        self._invalid_message = invalid_message
        self._rendered = False
        self._invalid = False
        self._loading = False
        self._application: Application
        self._long_instruction = long_instruction
        self._border = border
        self._height_offset = 2  # prev prompt result + current prompt question
        if self._border:
            self._height_offset += 2
        if self._long_instruction:
            self._height_offset += 1
        self._validation_window_bottom_offset = 0 if not self._long_instruction else 1
        if self._wrap_lines:
            self._validation_window_bottom_offset += (
                self.extra_long_instruction_line_count
            )

        self._is_vim_edit = Condition(lambda: self._editing_mode == EditingMode.VI)
        self._is_invalid = Condition(lambda: self._invalid)
        self._is_displaying_long_instruction = Condition(
            lambda: self._long_instruction != ""
        )

    def _redraw(self) -> None:
        """Redraw the application UI."""
        self._application.invalidate()

    def register_kb(
        self, *keys: Union[Keys, str], filter: FilterOrBool = True
    ) -> Callable[[KeyHandlerCallable], KeyHandlerCallable]:
        """Decorate keybinding registration function.

        Ensure that the `invalid` state is cleared on next keybinding entered.
        """
        kb_dec = super().register_kb(*keys, filter=filter)

        def decorator(func: KeyHandlerCallable) -> KeyHandlerCallable:
            @kb_dec
            def executable(event):
                if self._invalid:
                    self._invalid = False
                func(event)

            return executable

        return decorator

    def _exception_handler(self, _, context) -> None:
        """Set exception handler for the event loop.

        Skip the question and raise exception.

        Args:
            loop: Current event loop.
            context: Exception context.
        """
        self._status["answered"] = True
        self._status["result"] = INQUIRERPY_KEYBOARD_INTERRUPT
        self._status["skipped"] = True
        self._application.exit(exception=context["exception"])

    def _after_render(self, app: Optional[Application]) -> None:
        """Run after the :class:`~prompt_toolkit.application.Application` is rendered/updated.

        Since this function is fired up on each render, adding a check on `self._rendered` to
        process logics that should only run once.

        Set event loop exception handler here, since its guaranteed that the event loop is running
        in `_after_render`.
        """
        if not self._rendered:
            self._rendered = True

            self._keybinding_factory()
            self._on_rendered(app)

    def _set_error(self, message: str) -> None:
        """Set error message and set invalid state.

        Args:
            message: Error message to display.
        """
        self._invalid_message = message
        self._invalid = True

    def _get_error_message(self) -> List[Tuple[str, str]]:
        """Obtain the error message dynamically.

        Returns:
            FormattedText in list of tuple format.
        """
        return [
            (
                "class:validation-toolbar",
                self._invalid_message,
            )
        ]

    def _on_rendered(self, _: Optional[Application]) -> None:
        """Run once after the UI is rendered. Acts like `ComponentDidMount`."""
        pass

    def _get_prompt_message(self) -> List[Tuple[str, str]]:
        """Get the prompt message to display.

        Returns:
            Formatted text in list of tuple format.
        """
        pre_answer = (
            "class:instruction",
            " %s " % self.instruction if self.instruction else " ",
        )
        post_answer = ("class:answer", " %s" % self.status["result"])
        return super()._get_prompt_message(pre_answer, post_answer)

    def _run(self) -> Any:
        """Run the application."""
        return self.application.run()

    async def _run_async(self) -> None:
        """Run the application asynchronously."""
        return await self.application.run_async()

    @property
    def application(self) -> Application:
        """Get the application.

        :class:`.BaseComplexPrompt` requires :attr:`.BaseComplexPrompt._application` to be defined since this class
        doesn't implement :class:`~prompt_toolkit.layout.Layout` and :class:`~prompt_toolkit.application.Application`.

        Raises:
            NotImplementedError: When `self._application` is not defined.
        """
        if not self._application:
            raise NotImplementedError
        return self._application

    @application.setter
    def application(self, value: Application) -> None:
        self._application = value

    @property
    def height_offset(self) -> int:
        """int: Height offset to apply."""
        if not self._wrap_lines:
            return self._height_offset
        return self.extra_line_count + self._height_offset

    @property
    def total_message_length(self) -> int:
        """int: Total length of the message."""
        total_message_length = 0
        if self._qmark:
            total_message_length += len(self._qmark)
            total_message_length += 1  # Extra space if qmark is present
        total_message_length += len(str(self._message))
        total_message_length += 1  # Extra space between message and instruction
        total_message_length += len(str(self._instruction))
        if self._instruction:
            total_message_length += 1  # Extra space behind the instruction
        return total_message_length

    @property
    def extra_message_line_count(self) -> int:
        """int: Get the extra lines created caused by line wrapping.

        Minus 1 on the totoal message length as we only want the extra line.
        24 // 24 will equal to 1 however we only want the value to be 1 when we have 25 char
        which will create an extra line.
        """
        term_width, _ = shutil.get_terminal_size()
        return (self.total_message_length - 1) // term_width

    @property
    def extra_long_instruction_line_count(self) -> int:
        """int: Get the extra lines created caused by line wrapping.

        See Also:
            :attr:`.BaseComplexPrompt.extra_message_line_count`
        """
        if self._long_instruction:
            term_width, _ = shutil.get_terminal_size()
            return (len(self._long_instruction) - 1) // term_width
        else:
            return 0

    @property
    def extra_line_count(self) -> int:
        """Get the extra lines created caused by line wrapping.

        Used mainly to calculate how much additional offset should be applied when getting
        the height.

        Returns:
            Total extra lines created due to line wrapping.
        """
        result = 0

        # message wrap
        result += self.extra_message_line_count
        # long instruction wrap
        result += self.extra_long_instruction_line_count

        return result
